import unittest

from pypika import Case, Field, Query, Tables, Tuple, functions
from pypika.dialects import (
    ClickHouseQuery,
    ClickHouseQueryBuilder,
    MSSQLQuery,
    MSSQLQueryBuilder,
    MySQLLoadQueryBuilder,
    MySQLQuery,
    MySQLQueryBuilder,
    OracleQuery,
    OracleQueryBuilder,
    PostgreSQLQuery,
    PostgreSQLQueryBuilder,
    RedshiftQuery,
    RedShiftQueryBuilder,
    SnowflakeQuery,
    SnowflakeQueryBuilder,
    SQLLiteQuery,
    SQLLiteQueryBuilder,
    VerticaCopyQueryBuilder,
    VerticaCreateQueryBuilder,
    VerticaQuery,
    VerticaQueryBuilder,
)
from pypika.queries import CreateQueryBuilder, DropQueryBuilder, QueryBuilder


class QueryTablesTests(unittest.TestCase):
    table_a, table_b, table_c, table_d = Tables("a", "b", "c", "d")

    def test_replace_table(self):
        query = Query.from_(self.table_a).select(self.table_a.time)
        query = query.replace_table(self.table_a, self.table_b)

        self.assertEqual('SELECT "time" FROM "b"', str(query))

    def test_replace_only_specified_table(self):
        query = Query.from_(self.table_a).select(self.table_a.time)
        query = query.replace_table(self.table_b, self.table_c)

        self.assertEqual('SELECT "time" FROM "a"', str(query))

    def test_replace_insert_table(self):
        query = Query.into(self.table_a).insert(1)
        query = query.replace_table(self.table_a, self.table_b)

        self.assertEqual('INSERT INTO "b" VALUES (1)', str(query))

    def test_replace_insert_table_current_table_not_match(self):
        query = Query.into(self.table_a).insert(1)
        query = query.replace_table(self.table_c, self.table_b)

        self.assertEqual('INSERT INTO "a" VALUES (1)', str(query))

    def test_replace_update_table(self):
        query = Query.update(self.table_a).set("foo", "bar")
        query = query.replace_table(self.table_a, self.table_b)

        self.assertEqual('UPDATE "b" SET "foo"=\'bar\'', str(query))

    def test_replace_update_table_current_table_not_match(self):
        query = Query.update(self.table_a).set("foo", "bar")
        query = query.replace_table(self.table_c, self.table_b)

        self.assertEqual('UPDATE "a" SET "foo"=\'bar\'', str(query))

    def test_replace_delete_table(self):
        query = Query.from_(self.table_a).delete()
        query = query.replace_table(self.table_a, self.table_b)

        self.assertEqual('DELETE FROM "b"', str(query))

    def test_replace_join_tables(self):
        query = (
            Query.from_(self.table_a)
            .join(self.table_b)
            .on(self.table_a.customer_id == self.table_b.id)
            .join(self.table_c)
            .on(self.table_b.seller_id == self.table_c.id)
            .select(self.table_a.star)
        )
        query = query.replace_table(self.table_a, self.table_d)

        self.assertEqual(
            'SELECT "d".* FROM "d" JOIN "b" ON "d"."customer_id"="b"."id" JOIN "c" ON "b"."seller_id"="c"."id"',
            str(query),
        )

    def test_replace_filter_tables(self):
        query = Query.from_(self.table_a).select(self.table_a.name).where(self.table_a.name == "Mustermann")
        query = query.replace_table(self.table_a, self.table_b)

        self.assertEqual('SELECT "name" FROM "b" WHERE "name"=\'Mustermann\'', str(query))

    def test_replace_having_table(self):
        query = (
            Query.from_(self.table_a)
            .select(functions.Sum(self.table_a.revenue))
            .groupby(self.table_a.customer)
            .having(functions.Sum(self.table_a.revenue) >= 1000)
        )
        query = query.replace_table(self.table_a, self.table_b)

        self.assertEqual(
            'SELECT SUM("revenue") FROM "b" GROUP BY "customer" HAVING SUM("revenue")>=1000',
            str(query),
        )

    def test_replace_case_table(self):
        query = Query.from_(self.table_a).select(
            Case()
            .when(self.table_a.fname == "Tom", "It was Tom")
            .when(self.table_a.fname == "John", "It was John")
            .else_("It was someone else.")
            .as_("who_was_it"),
        )
        query = query.replace_table(self.table_a, self.table_b)

        self.assertEqual(
            "SELECT CASE "
            "WHEN \"fname\"='Tom' THEN 'It was Tom' "
            "WHEN \"fname\"='John' THEN 'It was John' "
            "ELSE 'It was someone else.' END \"who_was_it\" "
            'FROM "b"',
            str(query),
        )

    def test_replace_orderby_table(self):
        query = Query.from_(self.table_a).select(self.table_a.customer).orderby(self.table_a.customer)
        query = query.replace_table(self.table_a, self.table_b)

        self.assertEqual('SELECT "customer" FROM "b" ORDER BY "customer"', str(query))

    def test_replace_tuple_table(self):
        query = (
            Query.from_(self.table_a)
            .select(self.table_a.cost, self.table_a.revenue)
            .where((self.table_a.cost, self.table_a.revenue) == Tuple(1, 2))
        )

        query = query.replace_table(self.table_a, self.table_b)

        # Order is reversed due to lack of right equals method
        self.assertEqual(
            'SELECT "cost","revenue" FROM "b" WHERE (1,2)=("cost","revenue")',
            str(query),
        )

    def test_is_joined(self):
        q = Query.from_(self.table_a).join(self.table_b).on(self.table_a.foo == self.table_b.boo)

        self.assertTrue(q.is_joined(self.table_b))
        self.assertFalse(q.is_joined(self.table_c))


class QueryBuilderTests(unittest.TestCase):
    def test_query_builders_have_reference_to_correct_query_class(self):
        with self.subTest('QueryBuilder'):
            self.assertEqual(Query, QueryBuilder.QUERY_CLS)

        with self.subTest('DropQueryBuilder'):
            self.assertEqual(Query, DropQueryBuilder.QUERY_CLS)

        with self.subTest('CreateQueryBuilder'):
            self.assertEqual(Query, CreateQueryBuilder.QUERY_CLS)

        with self.subTest('MySQLQueryBuilder'):
            self.assertEqual(MySQLQuery, MySQLQueryBuilder.QUERY_CLS)

        with self.subTest('MySQLLoadQueryBuilder'):
            self.assertEqual(MySQLQuery, MySQLLoadQueryBuilder.QUERY_CLS)

        with self.subTest('VerticaQueryBuilder'):
            self.assertEqual(VerticaQuery, VerticaQueryBuilder.QUERY_CLS)

        with self.subTest('VerticaCreateQueryBuilder'):
            self.assertEqual(VerticaQuery, VerticaCreateQueryBuilder.QUERY_CLS)

        with self.subTest('VerticaCopyQueryBuilder'):
            self.assertEqual(VerticaQuery, VerticaCopyQueryBuilder.QUERY_CLS)

        with self.subTest('PostgreSQLQueryBuilder'):
            self.assertEqual(PostgreSQLQuery, PostgreSQLQueryBuilder.QUERY_CLS)

        with self.subTest('MSSQLQueryBuilder'):
            self.assertEqual(MSSQLQuery, MSSQLQueryBuilder.QUERY_CLS)

        with self.subTest('SnowflakeQueryBuilder'):
            self.assertEqual(SnowflakeQuery, SnowflakeQueryBuilder.QUERY_CLS)

        with self.subTest('ClickHouseQueryBuilder'):
            self.assertEqual(ClickHouseQuery, ClickHouseQueryBuilder.QUERY_CLS)

        with self.subTest('RedShiftQueryBuilder'):
            self.assertEqual(RedshiftQuery, RedShiftQueryBuilder.QUERY_CLS)

        with self.subTest('SQLLiteQueryBuilder'):
            self.assertEqual(SQLLiteQuery, SQLLiteQueryBuilder.QUERY_CLS)

        with self.subTest('OracleQueryBuilder'):
            self.assertEqual(OracleQuery, OracleQueryBuilder.QUERY_CLS)

    def test_pipe(self) -> None:
        base_query = Query.from_("test")

        def select(query: QueryBuilder) -> QueryBuilder:
            return query.select("test1", "test2")

        def count_group(query: QueryBuilder, *groups) -> QueryBuilder:
            return query.groupby(*groups).select(*groups, functions.Count("*"))

        for func, args, kwargs, expected_str in [
            (select, [], {}, 'SELECT "test1","test2" FROM "test"'),
            (
                count_group,
                ["test1", "test2"],
                {},
                'SELECT "test1","test2",COUNT(*) FROM "test" GROUP BY "test1","test2"',
            ),
            (count_group, ["test1"], {}, 'SELECT "test1",COUNT(*) FROM "test" GROUP BY "test1"'),
        ]:
            result_str = str(base_query.pipe(func, *args, **kwargs))
            self.assertEqual(result_str, str(func(base_query, *args, **kwargs)))
            self.assertEqual(result_str, expected_str)

        def where_clause(query: QueryBuilder, num_days: int) -> QueryBuilder:
            return query.where(Field("date") > functions.Now() - num_days)

        result_str = str(base_query.pipe(select).pipe(where_clause, num_days=1))
        self.assertEqual(
            result_str,
            str(select(where_clause(base_query, num_days=1))),
        )
        self.assertEqual(
            result_str,
            'SELECT "test1","test2" FROM "test" WHERE "date">NOW()-1',
        )



import unittest

from pypika import (
    AliasedQuery,
    Query,
    Tables,
)


class ImmutabilityTests(unittest.TestCase):
    table_a, table_b, table_c = Tables("a", "b", "c")

    def test_select_returns_new_query_instance(self):
        query_a = Query.from_(self.table_a).select(self.table_a.foo)
        query_b = query_a.select(self.table_a.bar)

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_groupby_returns_new_query_instance(self):
        query_a = Query.from_(self.table_a).select(self.table_a.foo).groupby(self.table_a.foo)
        query_b = query_a.groupby(self.table_a.bar)

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_orderby_return_new_query_instance(self):
        query_a = Query.from_(self.table_a).select(self.table_a.foo).orderby(self.table_a.foo)
        query_b = query_a.orderby(self.table_a.bar)

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_join_return_new_query_instance(self):
        base = Query.from_(self.table_a).select(self.table_a.foo)
        query_a = base.join(self.table_b).on(self.table_a.foo == self.table_b.bar).select(self.table_b.bar)
        query_b = query_a.join(self.table_c).on(self.table_a.foo == self.table_c.baz).select(self.table_c.baz)

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_use_index_return_new_query_instance(self):
        query_a = Query.from_(self.table_a).select(self.table_a.foo).use_index('idx1')
        query_b = query_a.use_index('idx2')

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_force_index_return_new_query_instance(self):
        query_a = Query.from_(self.table_a).select(self.table_a.foo).force_index('idx1')
        query_b = query_a.force_index('idx2')

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_with_return_new_query_instance(self):
        alias_1 = Query.from_(self.table_a).select(self.table_a.foo)
        alias_2 = Query.from_(self.table_b).select(self.table_b.bar)
        query_a = Query.from_(AliasedQuery('a1')).select('foo').with_(alias_1, 'a1')
        query_b = query_a.with_(alias_2, 'a2')

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_insert_into_return_new_query_instance(self):
        query_a = Query.into(self.table_a).insert('foo1')
        query_b = query_a.insert('foo2')

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_replace_into_return_new_query_instance(self):
        query_a = Query.into(self.table_a).replace('foo')
        query_b = query_a.replace('bar')

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_update_set_return_new_query_instance(self):
        query_a = Query.update(self.table_a).set(self.table_a.foo, 'foo')
        query_b = query_a.set(self.table_a.bar, 'bar')

        self.assertIsNot(query_a, query_b)
        self.assertNotEqual(str(query_a), str(query_b))

    def test_queries_after_join(self):
        query1 = Query.from_(self.table_a).select(self.table_a.foo)
        query2 = query1.join(self.table_b).on(self.table_a.foo == self.table_b.bar).select(self.table_b.buz)

        self.assertEqual('SELECT "foo" FROM "a"', str(query1))
        self.assertEqual(
            'SELECT "a"."foo","b"."buz" FROM "a" JOIN "b" ON "a"."foo"="b"."bar"',
            str(query2),
        )

    def test_immutable_kwarg_on_query_builder_disables_immutability(self):
        query0 = Query.from_(self.table_a, immutable=False)
        query1 = query0.select(self.table_a.foo)
        self.assertIs(query0, query1)



from collections.abc import Callable
from typing import Any, TypeVar

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class QueryException(Exception):
    pass


class GroupingException(Exception):
    pass


class CaseException(Exception):
    pass


class JoinException(Exception):
    pass


class SetOperationException(Exception):
    pass


class RollupException(Exception):
    pass


class DialectNotSupported(Exception):
    pass


class FunctionException(Exception):
    pass


C = TypeVar("C")


def builder(func: C) -> C:
    """
    Decorator for wrapper "builder" functions.  These are functions on the Query class or other classes used for
    building queries which mutate the query and return self.  To make the build functions immutable, this decorator is
    used which will deepcopy the current instance.  This decorator will return the return value of the inner function
    or the new copy of the instance.  The inner function does not need to return self.
    """
    import copy

    def _copy(self, *args, **kwargs):
        self_copy = copy.copy(self) if getattr(self, "immutable", True) else self
        result = func(self_copy, *args, **kwargs)

        # Return self if the inner function returns None.  This way the inner function can return something
        # different (for example when creating joins, a different builder is returned).
        if result is None:
            return self_copy

        return result

    return _copy


def ignore_copy(func: Callable) -> Callable:
    """
    Decorator for wrapping the __getattr__ function for classes that are copied via deepcopy.  This prevents infinite
    recursion caused by deepcopy looking for magic functions in the class. Any class implementing __getattr__ that is
    meant to be deepcopy'd should use this decorator.

    deepcopy is used by pypika in builder functions (decorated by @builder) to make the results immutable.  Any data
    model type class (stored in the Query instance) is copied.
    """

    def _getattr(self, name):
        if name in [
            "__copy__",
            "__deepcopy__",
            "__getstate__",
            "__setstate__",
            "__getnewargs__",
        ]:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

        return func(self, name)

    return _getattr


def resolve_is_aggregate(values: list[bool | None]) -> bool | None:
    """
    Resolves the is_aggregate flag for an expression that contains multiple terms.  This works like a voter system,
    each term votes True or False or abstains with None.

    :param values: A list of booleans (or None) for each term in the expression
    :return: If all values are True or None, True is returned.  If all values are None, None is returned. Otherwise,
        False is returned.
    """
    result = [x for x in values if x is not None]
    if result:
        return all(result)
    return None


def format_quotes(value: Any, quote_char: str | None) -> str:
    return "{quote}{value}{quote}".format(value=value, quote=quote_char or "")


def format_alias_sql(
    sql: str,
    alias: str | None,
    quote_char: str | None = None,
    alias_quote_char: str | None = None,
    as_keyword: bool = False,
    **kwargs: Any,
) -> str:
    if alias is None:
        return sql
    return "{sql}{_as}{alias}".format(
        sql=sql,
        _as=' AS ' if as_keyword else ' ',
        alias=format_quotes(alias, alias_quote_char or quote_char),
    )


def validate(*args: Any, exc: Exception | None = None, type: type | None = None) -> None:
    if type is not None:
        for arg in args:
            if not isinstance(arg, type):
                raise exc



import unittest

from pypika import Table
from pypika.analytics import Count
from pypika.dialects import MSSQLQuery
from pypika.utils import QueryException


class SelectTests(unittest.TestCase):
    def test_normal_select(self):
        q = MSSQLQuery.from_("abc").select("def")

        self.assertEqual('SELECT "def" FROM "abc"', str(q))

    def test_distinct_select(self):
        q = MSSQLQuery.from_("abc").select("def").distinct()

        self.assertEqual('SELECT DISTINCT "def" FROM "abc"', str(q))

    def test_top_distinct_select(self):
        q = MSSQLQuery.from_("abc").select("def").top(10).distinct()

        self.assertEqual('SELECT DISTINCT TOP (10) "def" FROM "abc"', str(q))

    def test_top_select(self):
        q = MSSQLQuery.from_("abc").select("def").top(10)

        self.assertEqual('SELECT TOP (10) "def" FROM "abc"', str(q))

    def test_top_with_ties_select(self):
        q = MSSQLQuery.from_("abc").select("def").top(10, with_ties=True)

        self.assertEqual('SELECT TOP (10) WITH TIES "def" FROM "abc"', str(q))

    def test_top_percent_select(self):
        q = MSSQLQuery.from_("abc").select("def").top(10, percent=True)

        self.assertEqual('SELECT TOP (10) PERCENT "def" FROM "abc"', str(q))

    def test_top_percent_invalid_range(self):
        for invalid_top in [-1, 101]:
            with self.assertRaisesRegex(
                QueryException,
                "TOP value must be between 0 and 100 when `percent` is specified",
            ):
                MSSQLQuery.from_("abc").select("def").top(invalid_top, percent=True)

    def test_top_select_non_int(self):
        with self.assertRaisesRegex(QueryException, "TOP value must be an integer"):
            MSSQLQuery.from_("abc").select("def").top("a")

    def test_limit(self):
        q = MSSQLQuery.from_("abc").select("def").orderby("def").limit(10)

        self.assertEqual('SELECT "def" FROM "abc" ORDER BY "def" OFFSET 0 ROWS FETCH NEXT 10 ROWS ONLY', str(q))

    def test_offset(self):
        q = MSSQLQuery.from_("abc").select("def").orderby("def").offset(10)

        self.assertEqual('SELECT "def" FROM "abc" ORDER BY "def" OFFSET 10 ROWS', str(q))

    def test_limit_with_offset(self):
        q = MSSQLQuery.from_("abc").select("def").orderby("def").limit(10).offset(10)

        self.assertEqual('SELECT "def" FROM "abc" ORDER BY "def" OFFSET 10 ROWS FETCH NEXT 10 ROWS ONLY', str(q))

    def test_groupby_alias_False_does_not_group_by_alias_with_standard_query(self):
        t = Table('table1')
        col = t.abc.as_('a')
        q = MSSQLQuery.from_(t).select(col, Count('*')).groupby(col)

        self.assertEqual('SELECT "abc" "a",COUNT(\'*\') FROM "table1" GROUP BY "abc"', str(q))

    def test_groupby_alias_False_does_not_group_by_alias_when_subqueries_are_present(self):
        t = Table('table1')
        subquery = MSSQLQuery.from_(t).select(t.abc)
        col = subquery.abc.as_('a')
        q = MSSQLQuery.from_(subquery).select(col, Count('*')).groupby(col)

        self.assertEqual(
            'SELECT "sq0"."abc" "a",COUNT(\'*\') FROM (SELECT "abc" FROM "table1") "sq0" GROUP BY "sq0"."abc"',
            str(q),
        )



import unittest

from pypika import Column, MySQLQuery, QueryException, Table


class SelectTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_normal_select(self):
        q = MySQLQuery.from_("abc").select("def")

        self.assertEqual("SELECT `def` FROM `abc`", str(q))

    def test_distinct_select(self):
        q = MySQLQuery.from_("abc").select("def").distinct()

        self.assertEqual("SELECT DISTINCT `def` FROM `abc`", str(q))

    def test_modifier_select(self):
        q = MySQLQuery.from_("abc").select("def").select("ghi").modifier("SQL_CALC_FOUND_ROWS")

        self.assertEqual("SELECT SQL_CALC_FOUND_ROWS `def`,`ghi` FROM `abc`", str(q))

    def test_multiple_modifier_select(self):
        q = MySQLQuery.from_("abc").select("def").modifier("HIGH_PRIORITY").modifier("SQL_CALC_FOUND_ROWS")

        self.assertEqual("SELECT HIGH_PRIORITY SQL_CALC_FOUND_ROWS `def` FROM `abc`", str(q))


class UpdateTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_update(self):
        q = MySQLQuery.into("abc").insert(1, [1, "a", True])

        self.assertEqual("INSERT INTO `abc` VALUES (1,[1,'a',true])", str(q))

    def test_on_duplicate_key_ignore_update(self):
        q = MySQLQuery.into("abc").insert(1, [1, "a", True]).on_duplicate_key_ignore()

        self.assertEqual("INSERT INTO `abc` VALUES (1,[1,'a',true]) ON DUPLICATE KEY IGNORE", str(q))

    def test_on_duplicate_key_update_update(self):
        q = MySQLQuery.into("abc").insert(1, [1, "a", True]).on_duplicate_key_update(self.table_abc.a, 'b')

        self.assertEqual("INSERT INTO `abc` VALUES (1,[1,'a',true]) ON DUPLICATE KEY UPDATE `a`='b'", str(q))

    def test_conflict_handlers_update(self):
        with self.assertRaises(QueryException):
            (
                MySQLQuery.into("abc")
                .insert(1, [1, "a", True])
                .on_duplicate_key_ignore()
                .on_duplicate_key_update(self.table_abc.a, 'b')
            )


class LoadCSVTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_load_from_file(self):
        q1 = MySQLQuery.load("/path/to/file").into("abc")

        q2 = MySQLQuery.load("/path/to/file").into(self.table_abc)

        self.assertEqual(
            "LOAD DATA LOCAL INFILE '/path/to/file' INTO TABLE `abc` FIELDS TERMINATED BY ','",
            str(q1),
        )
        self.assertEqual(
            "LOAD DATA LOCAL INFILE '/path/to/file' INTO TABLE `abc` FIELDS TERMINATED BY ','",
            str(q2),
        )


class TableTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_create_table(self):
        q = MySQLQuery.create_table(self.table_abc).columns(Column("id", "INT"))
        self.assertEqual(
            'CREATE TABLE `abc` (`id` INT)',
            str(q),
        )

    def test_drop_table(self):
        q = MySQLQuery.drop_table(self.table_abc)
        self.assertEqual(
            'DROP TABLE `abc`',
            str(q),
        )



import itertools
import warnings
from copy import copy
from typing import Any

from pypika.enums import Dialects
from pypika.queries import (
    CreateQueryBuilder,
    Database,
    DropQueryBuilder,
    Query,
    QueryBuilder,
    Selectable,
    Table,
)
from pypika.terms import ArithmeticExpression, Criterion, EmptyCriterion, Field, Function, Star, Term, ValueWrapper
from pypika.utils import QueryException, builder, format_quotes


class SnowflakeQuery(Query):
    """
    Defines a query class for use with Snowflake.
    """

    @classmethod
    def _builder(cls, **kwargs: Any) -> "SnowflakeQueryBuilder":
        return SnowflakeQueryBuilder(**kwargs)

    @classmethod
    def create_table(cls, table: str | Table) -> "SnowflakeCreateQueryBuilder":
        return SnowflakeCreateQueryBuilder().create_table(table)

    @classmethod
    def drop_table(cls, table: str | Table) -> "SnowflakeDropQueryBuilder":
        return SnowflakeDropQueryBuilder().drop_table(table)


class SnowflakeQueryBuilder(QueryBuilder):
    QUOTE_CHAR = None
    ALIAS_QUOTE_CHAR = '"'
    QUERY_ALIAS_QUOTE_CHAR = ''
    QUERY_CLS = SnowflakeQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=Dialects.SNOWFLAKE, **kwargs)


class SnowflakeCreateQueryBuilder(CreateQueryBuilder):
    QUOTE_CHAR = None
    QUERY_CLS = SnowflakeQuery

    def __init__(self) -> None:
        super().__init__(dialect=Dialects.SNOWFLAKE)


class SnowflakeDropQueryBuilder(DropQueryBuilder):
    QUOTE_CHAR = None
    QUERY_CLS = SnowflakeQuery

    def __init__(self) -> None:
        super().__init__(dialect=Dialects.SNOWFLAKE)


class MySQLQuery(Query):
    """
    Defines a query class for use with MySQL.
    """

    @classmethod
    def _builder(cls, **kwargs: Any) -> "MySQLQueryBuilder":
        return MySQLQueryBuilder(**kwargs)

    @classmethod
    def load(cls, fp: str) -> "MySQLLoadQueryBuilder":
        return MySQLLoadQueryBuilder().load(fp)

    @classmethod
    def create_table(cls, table: str | Table) -> "MySQLCreateQueryBuilder":
        return MySQLCreateQueryBuilder().create_table(table)

    @classmethod
    def drop_table(cls, table: str | Table) -> "MySQLDropQueryBuilder":
        return MySQLDropQueryBuilder().drop_table(table)


class MySQLQueryBuilder(QueryBuilder):
    QUOTE_CHAR = "`"
    QUERY_CLS = MySQLQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=Dialects.MYSQL, **kwargs)
        self._duplicate_updates = []
        self._ignore_duplicates = False
        self._modifiers = []

        self._for_update_nowait = False
        self._for_update_skip_locked = False
        self._for_update_of = set()

    def __copy__(self) -> "MySQLQueryBuilder":
        newone = super().__copy__()
        newone._duplicate_updates = copy(self._duplicate_updates)
        newone._ignore_duplicates = copy(self._ignore_duplicates)
        return newone

    @builder
    def for_update(
        self,
        nowait: bool = False,
        skip_locked: bool = False,
        of: tuple[str, ...] = (),
    ) -> "QueryBuilder":
        self._for_update = True
        self._for_update_skip_locked = skip_locked
        self._for_update_nowait = nowait
        self._for_update_of = set(of)

    @builder
    def on_duplicate_key_update(self, field: Field | str, value: Any) -> "MySQLQueryBuilder":
        if self._ignore_duplicates:
            raise QueryException("Can not have two conflict handlers")

        field = Field(field) if not isinstance(field, Field) else field
        self._duplicate_updates.append((field, ValueWrapper(value)))

    @builder
    def on_duplicate_key_ignore(self) -> "MySQLQueryBuilder":
        if self._duplicate_updates:
            raise QueryException("Can not have two conflict handlers")

        self._ignore_duplicates = True

    def get_sql(self, **kwargs: Any) -> str:
        self._set_kwargs_defaults(kwargs)
        querystring = super(MySQLQueryBuilder, self).get_sql(**kwargs)
        if querystring:
            if self._duplicate_updates:
                querystring += self._on_duplicate_key_update_sql(**kwargs)
            elif self._ignore_duplicates:
                querystring += self._on_duplicate_key_ignore_sql()
        return querystring

    def _for_update_sql(self, **kwargs) -> str:
        if self._for_update:
            for_update = ' FOR UPDATE'
            if self._for_update_of:
                for_update += f' OF {", ".join([Table(item).get_sql(**kwargs) for item in self._for_update_of])}'
            if self._for_update_nowait:
                for_update += ' NOWAIT'
            elif self._for_update_skip_locked:
                for_update += ' SKIP LOCKED'
        else:
            for_update = ''

        return for_update

    def _on_duplicate_key_update_sql(self, **kwargs: Any) -> str:
        return " ON DUPLICATE KEY UPDATE {updates}".format(
            updates=",".join(
                f"{field.get_sql(**kwargs)}={value.get_sql(**kwargs)}" for field, value in self._duplicate_updates
            ),
        )

    def _on_duplicate_key_ignore_sql(self) -> str:
        return " ON DUPLICATE KEY IGNORE"

    @builder
    def modifier(self, value: str) -> "MySQLQueryBuilder":
        """
        Adds a modifier such as SQL_CALC_FOUND_ROWS to the query.
        https://dev.mysql.com/doc/refman/5.7/en/select.html

        :param value: The modifier value e.g. SQL_CALC_FOUND_ROWS
        """
        self._modifiers.append(value)

    def _select_sql(self, **kwargs: Any) -> str:
        """
        Overridden function to generate the SELECT part of the SQL statement,
        with the addition of the a modifier if present.
        """
        return "SELECT {distinct}{modifier}{select}".format(
            distinct="DISTINCT " if self._distinct else "",
            modifier="{} ".format(" ".join(self._modifiers)) if self._modifiers else "",
            select=",".join(term.get_sql(with_alias=True, subquery=True, **kwargs) for term in self._selects),
        )


class MySQLLoadQueryBuilder:
    QUERY_CLS = MySQLQuery

    def __init__(self) -> None:
        self._load_file = None
        self._into_table = None

    @builder
    def load(self, fp: str) -> "MySQLLoadQueryBuilder":
        self._load_file = fp

    @builder
    def into(self, table: str | Table) -> "MySQLLoadQueryBuilder":
        self._into_table = table if isinstance(table, Table) else Table(table)

    def get_sql(self, *args: Any, **kwargs: Any) -> str:
        querystring = ""
        if self._load_file and self._into_table:
            querystring += self._load_file_sql(**kwargs)
            querystring += self._into_table_sql(**kwargs)
            querystring += self._options_sql(**kwargs)

        return querystring

    def _load_file_sql(self, **kwargs: Any) -> str:
        return f"LOAD DATA LOCAL INFILE '{self._load_file}'"

    def _into_table_sql(self, **kwargs: Any) -> str:
        return f" INTO TABLE `{self._into_table.get_sql(**kwargs)}`"

    def _options_sql(self, **kwargs: Any) -> str:
        return " FIELDS TERMINATED BY ','"

    def __str__(self) -> str:
        return self.get_sql()


class MySQLCreateQueryBuilder(CreateQueryBuilder):
    QUOTE_CHAR = "`"


class MySQLDropQueryBuilder(DropQueryBuilder):
    QUOTE_CHAR = "`"


class VerticaQuery(Query):
    """
    Defines a query class for use with Vertica.
    """

    @classmethod
    def _builder(cls, **kwargs) -> "VerticaQueryBuilder":
        return VerticaQueryBuilder(**kwargs)

    @classmethod
    def from_file(cls, fp: str) -> "VerticaCopyQueryBuilder":
        return VerticaCopyQueryBuilder().from_file(fp)

    @classmethod
    def create_table(cls, table: str | Table) -> "VerticaCreateQueryBuilder":
        return VerticaCreateQueryBuilder().create_table(table)


class VerticaQueryBuilder(QueryBuilder):
    QUERY_CLS = VerticaQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=Dialects.VERTICA, **kwargs)
        self._hint = None

    @builder
    def hint(self, label: str) -> "VerticaQueryBuilder":
        self._hint = label

    def get_sql(self, *args: Any, **kwargs: Any) -> str:
        sql = super().get_sql(*args, **kwargs)

        if self._hint is not None:
            sql = "".join([sql[:7], f"/*+label({self._hint})*/", sql[6:]])

        return sql


class VerticaCreateQueryBuilder(CreateQueryBuilder):
    QUERY_CLS = VerticaQuery

    def __init__(self) -> None:
        super().__init__(dialect=Dialects.VERTICA)
        self._local = False
        self._preserve_rows = False

    @builder
    def local(self) -> "VerticaCreateQueryBuilder":
        if not self._temporary:
            raise AttributeError("'Query' object has no attribute temporary")

        self._local = True

    @builder
    def preserve_rows(self) -> "VerticaCreateQueryBuilder":
        if not self._temporary:
            raise AttributeError("'Query' object has no attribute temporary")

        self._preserve_rows = True

    def _create_table_sql(self, **kwargs: Any) -> str:
        return "CREATE {local}{temporary}TABLE {table}".format(
            local="LOCAL " if self._local else "",
            temporary="TEMPORARY " if self._temporary else "",
            table=self._create_table.get_sql(**kwargs),
        )

    def _table_options_sql(self, **kwargs) -> str:
        table_options = super()._table_options_sql(**kwargs)
        table_options += self._preserve_rows_sql()
        return table_options

    def _as_select_sql(self, **kwargs: Any) -> str:
        return f"{self._preserve_rows_sql()} AS ({self._as_select.get_sql(**kwargs)})"

    def _preserve_rows_sql(self) -> str:
        return " ON COMMIT PRESERVE ROWS" if self._preserve_rows else ""


class VerticaCopyQueryBuilder:
    QUERY_CLS = VerticaQuery

    def __init__(self) -> None:
        self._copy_table = None
        self._from_file = None

    @builder
    def from_file(self, fp: str) -> "VerticaCopyQueryBuilder":
        self._from_file = fp

    @builder
    def copy_(self, table: str | Table) -> "VerticaCopyQueryBuilder":
        self._copy_table = table if isinstance(table, Table) else Table(table)

    def get_sql(self, *args: Any, **kwargs: Any) -> str:
        querystring = ""
        if self._copy_table and self._from_file:
            querystring += self._copy_table_sql(**kwargs)
            querystring += self._from_file_sql(**kwargs)
            querystring += self._options_sql(**kwargs)

        return querystring

    def _copy_table_sql(self, **kwargs: Any) -> str:
        return f'COPY "{self._copy_table.get_sql(**kwargs)}"'

    def _from_file_sql(self, **kwargs: Any) -> str:
        return f" FROM LOCAL '{self._from_file}'"

    def _options_sql(self, **kwargs: Any) -> str:
        return " PARSER fcsvparser(header=false)"

    def __str__(self) -> str:
        return self.get_sql()


class FetchNextAndOffsetRowsQueryBuilder(QueryBuilder):
    def _limit_sql(self) -> str:
        return f" FETCH NEXT {self._limit} ROWS ONLY"

    def _offset_sql(self) -> str:
        return f" OFFSET {self._offset or 0} ROWS"

    @builder
    def fetch_next(self, limit: int):
        warnings.warn("`fetch_next` is deprecated - please use the `limit` method", DeprecationWarning)
        self._limit = limit


class OracleQuery(Query):
    """
    Defines a query class for use with Oracle.
    """

    @classmethod
    def _builder(cls, **kwargs: Any) -> "OracleQueryBuilder":
        return OracleQueryBuilder(**kwargs)


class OracleQueryBuilder(FetchNextAndOffsetRowsQueryBuilder):
    QUOTE_CHAR = None
    QUERY_CLS = OracleQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=Dialects.ORACLE, **kwargs)

    def get_sql(self, *args: Any, **kwargs: Any) -> str:
        # Oracle does not support group by a field alias
        # Note: set directly in kwargs as they are re-used down the tree in the case of subqueries!
        kwargs['groupby_alias'] = False
        return super().get_sql(*args, **kwargs)

    def _apply_pagination(self, querystring: str, **kwargs) -> str:
        # Note: Overridden as Oracle specifies offset before the fetch next limit
        if self._offset:
            querystring += self._offset_sql()

        if self._limit is not None:
            querystring += self._limit_sql()

        return querystring


class PostgreSQLQuery(Query):
    """
    Defines a query class for use with PostgreSQL.
    """

    @classmethod
    def _builder(cls, **kwargs) -> "PostgreSQLQueryBuilder":
        return PostgreSQLQueryBuilder(**kwargs)


class PostgreSQLQueryBuilder(QueryBuilder):
    ALIAS_QUOTE_CHAR = '"'
    QUERY_CLS = PostgreSQLQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=Dialects.POSTGRESQL, **kwargs)
        self._returns = []
        self._return_star = False

        self._on_conflict = False
        self._on_conflict_fields = []
        self._on_conflict_do_nothing = False
        self._on_conflict_do_updates = []
        self._on_conflict_wheres = None
        self._on_conflict_do_update_wheres = None

        self._distinct_on = []

        self._for_update_nowait = False
        self._for_update_skip_locked = False
        self._for_update_of = set()

    def __copy__(self) -> "PostgreSQLQueryBuilder":
        newone = super().__copy__()
        newone._returns = copy(self._returns)
        newone._on_conflict_do_updates = copy(self._on_conflict_do_updates)
        return newone

    @builder
    def distinct_on(self, *fields: str | Term) -> "PostgreSQLQueryBuilder":
        for field in fields:
            if isinstance(field, str):
                self._distinct_on.append(Field(field))
            elif isinstance(field, Term):
                self._distinct_on.append(field)

    @builder
    def for_update(
        self,
        nowait: bool = False,
        skip_locked: bool = False,
        of: tuple[str, ...] = (),
    ) -> "QueryBuilder":
        self._for_update = True
        self._for_update_skip_locked = skip_locked
        self._for_update_nowait = nowait
        self._for_update_of = set(of)

    @builder
    def on_conflict(self, *target_fields: str | Term) -> "PostgreSQLQueryBuilder":
        if not self._insert_table:
            raise QueryException("On conflict only applies to insert query")

        self._on_conflict = True

        for target_field in target_fields:
            if isinstance(target_field, str):
                self._on_conflict_fields.append(self._conflict_field_str(target_field))
            elif isinstance(target_field, Term):
                self._on_conflict_fields.append(target_field)

    @builder
    def do_nothing(self) -> "PostgreSQLQueryBuilder":
        if len(self._on_conflict_do_updates) > 0:
            raise QueryException("Can not have two conflict handlers")
        self._on_conflict_do_nothing = True

    @builder
    def do_update(
        self,
        update_field: str | Field,
        update_value: Any | None = None,
    ) -> "PostgreSQLQueryBuilder":
        if self._on_conflict_do_nothing:
            raise QueryException("Can not have two conflict handlers")

        if isinstance(update_field, str):
            field = self._conflict_field_str(update_field)
        elif isinstance(update_field, Field):
            field = update_field
        else:
            raise QueryException("Unsupported update_field")

        if update_value is not None:
            self._on_conflict_do_updates.append((field, ValueWrapper(update_value)))
        else:
            self._on_conflict_do_updates.append((field, None))

    @builder
    def where(self, criterion: Criterion) -> "PostgreSQLQueryBuilder":
        if not self._on_conflict:
            return super().where(criterion)

        if isinstance(criterion, EmptyCriterion):
            return

        if self._on_conflict_do_nothing:
            raise QueryException('DO NOTHING doest not support WHERE')

        if self._on_conflict_fields and self._on_conflict_do_updates:
            if self._on_conflict_do_update_wheres:
                self._on_conflict_do_update_wheres &= criterion
            else:
                self._on_conflict_do_update_wheres = criterion
        elif self._on_conflict_fields:
            if self._on_conflict_wheres:
                self._on_conflict_wheres &= criterion
            else:
                self._on_conflict_wheres = criterion
        else:
            raise QueryException('Can not have fieldless ON CONFLICT WHERE')

    @builder
    def using(self, table: Selectable | str) -> "QueryBuilder":
        self._using.append(table)

    def _distinct_sql(self, **kwargs: Any) -> str:
        if self._distinct_on:
            return "DISTINCT ON({distinct_on}) ".format(
                distinct_on=",".join(term.get_sql(with_alias=True, **kwargs) for term in self._distinct_on),
            )
        return super()._distinct_sql(**kwargs)

    def _conflict_field_str(self, term: str) -> Field | None:
        if self._insert_table:
            return Field(term, table=self._insert_table)

    def _on_conflict_sql(self, **kwargs: Any) -> str:
        if not self._on_conflict_do_nothing and len(self._on_conflict_do_updates) == 0:
            if not self._on_conflict_fields:
                return ""
            raise QueryException("No handler defined for on conflict")

        if self._on_conflict_do_updates and not self._on_conflict_fields:
            raise QueryException("Can not have fieldless on conflict do update")

        conflict_query = " ON CONFLICT"
        if self._on_conflict_fields:
            fields = [f.get_sql(with_alias=True, **kwargs) for f in self._on_conflict_fields]
            conflict_query += " (" + ', '.join(fields) + ")"

        if self._on_conflict_wheres:
            conflict_query += f" WHERE {self._on_conflict_wheres.get_sql(subquery=True, **kwargs)}"

        return conflict_query

    def _for_update_sql(self, **kwargs) -> str:
        if self._for_update:
            for_update = ' FOR UPDATE'
            if self._for_update_of:
                for_update += f' OF {", ".join([Table(item).get_sql(**kwargs) for item in self._for_update_of])}'
            if self._for_update_nowait:
                for_update += ' NOWAIT'
            elif self._for_update_skip_locked:
                for_update += ' SKIP LOCKED'
        else:
            for_update = ''

        return for_update

    def _on_conflict_action_sql(self, **kwargs: Any) -> str:
        if self._on_conflict_do_nothing:
            return " DO NOTHING"
        elif len(self._on_conflict_do_updates) > 0:
            updates = []
            for field, value in self._on_conflict_do_updates:
                if value:
                    updates.append(
                        f"{field.get_sql(**kwargs)}={value.get_sql(with_namespace=True, **kwargs)}",
                    )
                else:
                    updates.append(
                        f"{field.get_sql(**kwargs)}=EXCLUDED.{field.get_sql(**kwargs)}",
                    )
            action_sql = " DO UPDATE SET {updates}".format(updates=",".join(updates))

            if self._on_conflict_do_update_wheres:
                action_sql += (
                    f" WHERE {self._on_conflict_do_update_wheres.get_sql(subquery=True, with_namespace=True, **kwargs)}"
                )
            return action_sql

        return ''

    @builder
    def returning(self, *terms: Any) -> "PostgreSQLQueryBuilder":
        for term in terms:
            if isinstance(term, Field):
                self._return_field(term)
            elif isinstance(term, str):
                self._return_field_str(term)
            elif isinstance(term, (Function, ArithmeticExpression)):
                if term.is_aggregate:
                    raise QueryException("Aggregate functions are not allowed in returning")
                self._return_other(term)
            else:
                self._return_other(self.wrap_constant(term, self._wrapper_cls))

    def _validate_returning_term(self, term: Term) -> None:
        for field in term.fields_():
            if not any([self._insert_table, self._update_table, self._delete_from]):
                raise QueryException("Returning can't be used in this query")

            table_is_insert_or_update_table = field.table in {self._insert_table, self._update_table}
            join_tables = set(itertools.chain.from_iterable([j.criterion.tables_ for j in self._joins]))
            join_and_base_tables = set(self._from) | join_tables
            table_not_base_or_join = bool(term.tables_ - join_and_base_tables)
            if not table_is_insert_or_update_table and table_not_base_or_join:
                raise QueryException("You can't return from other tables")

    def _set_returns_for_star(self) -> None:
        self._returns = [returning for returning in self._returns if not hasattr(returning, "table")]
        self._return_star = True

    def _return_field(self, term: str | Field) -> None:
        if self._return_star:
            # Do not add select terms after a star is selected
            return

        self._validate_returning_term(term)

        if isinstance(term, Star):
            self._set_returns_for_star()

        self._returns.append(term)

    def _return_field_str(self, term: str | Field) -> None:
        if term == "*":
            self._set_returns_for_star()
            self._returns.append(Star())
            return

        if self._insert_table:
            self._return_field(Field(term, table=self._insert_table))
        elif self._update_table:
            self._return_field(Field(term, table=self._update_table))
        elif self._delete_from:
            self._return_field(Field(term, table=self._from[0]))
        else:
            raise QueryException("Returning can't be used in this query")

    def _return_other(self, function: Term) -> None:
        self._validate_returning_term(function)
        self._returns.append(function)

    def _returning_sql(self, **kwargs: Any) -> str:
        return " RETURNING {returning}".format(
            returning=",".join(term.get_sql(with_alias=True, **kwargs) for term in self._returns),
        )

    def get_sql(self, with_alias: bool = False, subquery: bool = False, **kwargs: Any) -> str:
        self._set_kwargs_defaults(kwargs)

        querystring = super(PostgreSQLQueryBuilder, self).get_sql(with_alias, subquery, **kwargs)

        querystring += self._on_conflict_sql(**kwargs)
        querystring += self._on_conflict_action_sql(**kwargs)

        if self._returns:
            kwargs['with_namespace'] = self._update_table and self.from_
            querystring += self._returning_sql(**kwargs)
        return querystring


class RedshiftQuery(Query):
    """
    Defines a query class for use with Amazon Redshift.
    """

    @classmethod
    def _builder(cls, **kwargs: Any) -> "RedShiftQueryBuilder":
        return RedShiftQueryBuilder(dialect=Dialects.REDSHIFT, **kwargs)


class RedShiftQueryBuilder(QueryBuilder):
    QUERY_CLS = RedshiftQuery


class MSSQLQuery(Query):
    """
    Defines a query class for use with Microsoft SQL Server.
    """

    @classmethod
    def _builder(cls, **kwargs: Any) -> "MSSQLQueryBuilder":
        return MSSQLQueryBuilder(**kwargs)


class MSSQLQueryBuilder(FetchNextAndOffsetRowsQueryBuilder):
    QUERY_CLS = MSSQLQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=Dialects.MSSQL, **kwargs)
        self._top: int | None = None
        self._top_with_ties: bool = False
        self._top_percent: bool = False

    @builder
    def top(self, value: str | int, percent: bool = False, with_ties: bool = False) -> "MSSQLQueryBuilder":
        """
        Implements support for simple TOP clauses.
        https://docs.microsoft.com/en-us/sql/t-sql/queries/top-transact-sql?view=sql-server-2017
        """
        try:
            self._top = int(value)
        except ValueError:
            raise QueryException("TOP value must be an integer")

        if percent and not (0 <= int(value) <= 100):
            raise QueryException("TOP value must be between 0 and 100 when `percent` is specified")
        self._top_percent: bool = percent
        self._top_with_ties: bool = with_ties

    def _apply_pagination(self, querystring: str, **kwargs) -> str:
        # Note: Overridden as MSSQL specifies offset before the fetch next limit
        if self._limit is not None or self._offset:
            # Offset has to be present if fetch next is specified in a MSSQL query
            querystring += self._offset_sql()

        if self._limit is not None:
            querystring += self._limit_sql()

        return querystring

    def get_sql(self, *args: Any, **kwargs: Any) -> str:
        # MSSQL does not support group by a field alias.
        # Note: set directly in kwargs as they are re-used down the tree in the case of subqueries!
        kwargs['groupby_alias'] = False
        return super().get_sql(*args, **kwargs)

    def _top_sql(self) -> str:
        _top_statement: str = ""
        if self._top:
            _top_statement = f"TOP ({self._top}) "
            if self._top_percent:
                _top_statement = f"{_top_statement}PERCENT "
            if self._top_with_ties:
                _top_statement = f"{_top_statement}WITH TIES "

        return _top_statement

    def _select_sql(self, **kwargs: Any) -> str:
        return "SELECT {distinct}{top}{select}".format(
            top=self._top_sql(),
            distinct="DISTINCT " if self._distinct else "",
            select=",".join(term.get_sql(with_alias=True, subquery=True, **kwargs) for term in self._selects),
        )


class ClickHouseQuery(Query):
    """
    Defines a query class for use with Yandex ClickHouse.
    """

    @classmethod
    def _builder(cls, **kwargs: Any) -> "ClickHouseQueryBuilder":
        return ClickHouseQueryBuilder(
            dialect=Dialects.CLICKHOUSE,
            wrap_set_operation_queries=False,
            as_keyword=True,
            **kwargs,
        )

    @classmethod
    def drop_database(self, database: Database | str) -> "ClickHouseDropQueryBuilder":
        return ClickHouseDropQueryBuilder().drop_database(database)

    @classmethod
    def drop_table(self, table: Table | str) -> "ClickHouseDropQueryBuilder":
        return ClickHouseDropQueryBuilder().drop_table(table)

    @classmethod
    def drop_dictionary(self, dictionary: str) -> "ClickHouseDropQueryBuilder":
        return ClickHouseDropQueryBuilder().drop_dictionary(dictionary)

    @classmethod
    def drop_quota(self, quota: str) -> "ClickHouseDropQueryBuilder":
        return ClickHouseDropQueryBuilder().drop_quota(quota)

    @classmethod
    def drop_user(self, user: str) -> "ClickHouseDropQueryBuilder":
        return ClickHouseDropQueryBuilder().drop_user(user)

    @classmethod
    def drop_view(self, view: str) -> "ClickHouseDropQueryBuilder":
        return ClickHouseDropQueryBuilder().drop_view(view)


class ClickHouseQueryBuilder(QueryBuilder):
    QUERY_CLS = ClickHouseQuery

    _distinct_on: list[Term]
    _limit_by: tuple[int, int, list[Term]] | None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._final = False
        self._sample = None
        self._sample_offset = None
        self._distinct_on = []
        self._limit_by = None

    def __copy__(self) -> "ClickHouseQueryBuilder":
        newone = super().__copy__()
        newone._limit_by = copy(self._limit_by)
        return newone

    @builder
    def final(self) -> "ClickHouseQueryBuilder":
        self._final = True

    @builder
    def sample(self, sample: int, offset: int | None = None) -> "ClickHouseQueryBuilder":
        self._sample = sample
        self._sample_offset = offset

    @staticmethod
    def _delete_sql(**kwargs: Any) -> str:
        return 'ALTER TABLE'

    def _update_sql(self, **kwargs: Any) -> str:
        return f"ALTER TABLE {self._update_table.get_sql(**kwargs)}"

    def _from_sql(self, with_namespace: bool = False, **kwargs: Any) -> str:
        selectable = ",".join(clause.get_sql(subquery=True, with_alias=True, **kwargs) for clause in self._from)
        if self._delete_from:
            return f" {selectable} DELETE"
        clauses = [selectable]
        if self._final is not False:
            clauses.append("FINAL")
        if self._sample is not None:
            clauses.append(f"SAMPLE {self._sample}")
        if self._sample_offset is not None:
            clauses.append(f"OFFSET {self._sample_offset}")
        return " FROM {clauses}".format(clauses=" ".join(clauses))

    def _set_sql(self, **kwargs: Any) -> str:
        return " UPDATE {set}".format(
            set=",".join(
                f"{field.get_sql(**dict(kwargs, with_namespace=False))}={value.get_sql(**kwargs)}"
                for field, value in self._updates
            ),
        )

    @builder
    def distinct_on(self, *fields: str | Term) -> "ClickHouseQueryBuilder":
        for field in fields:
            if isinstance(field, str):
                self._distinct_on.append(Field(field))
            elif isinstance(field, Term):
                self._distinct_on.append(field)

    def _distinct_sql(self, **kwargs: Any) -> str:
        if self._distinct_on:
            return "DISTINCT ON({distinct_on}) ".format(
                distinct_on=",".join(term.get_sql(with_alias=True, **kwargs) for term in self._distinct_on),
            )
        return super()._distinct_sql(**kwargs)

    @builder
    def limit_by(self, n, *by: str | Term) -> "ClickHouseQueryBuilder":
        self._limit_by = (n, 0, [Field(field) if isinstance(field, str) else field for field in by])

    @builder
    def limit_offset_by(self, n, offset, *by: str | Term) -> "ClickHouseQueryBuilder":
        self._limit_by = (n, offset, [Field(field) if isinstance(field, str) else field for field in by])

    def _apply_pagination(self, querystring: str, **kwargs) -> str:
        # LIMIT BY isn't really a pagination per se but since we need
        # to add this to the query right before an actual LIMIT clause
        # this is good enough.
        if self._limit_by:
            querystring += self._limit_by_sql(**kwargs)
        return super()._apply_pagination(querystring, **kwargs)

    def _limit_by_sql(self, **kwargs: Any) -> str:
        (n, offset, by) = self._limit_by
        by = ",".join(term.get_sql(with_alias=True, **kwargs) for term in by)
        if offset != 0:
            return f" LIMIT {n} OFFSET {offset} BY ({by})"
        else:
            return f" LIMIT {n} BY ({by})"

    def replace_table(self, current_table: Table | None, new_table: Table | None) -> "ClickHouseQueryBuilder":
        newone = super().replace_table(current_table, new_table)
        if self._limit_by:
            newone._limit_by = (
                self._limit_by[0],
                self._limit_by[1],
                [column.replace_table(current_table, new_table) for column in self._limit_by[2]],
            )
        return newone


class ClickHouseDropQueryBuilder(DropQueryBuilder):
    QUERY_CLS = ClickHouseQuery

    def __init__(self):
        super().__init__(dialect=Dialects.CLICKHOUSE)
        self._cluster_name = None

    @builder
    def drop_dictionary(self, dictionary: str) -> "ClickHouseDropQueryBuilder":
        super()._set_target('DICTIONARY', dictionary)

    @builder
    def drop_quota(self, quota: str) -> "ClickHouseDropQueryBuilder":
        super()._set_target('QUOTA', quota)

    @builder
    def on_cluster(self, cluster: str) -> "ClickHouseDropQueryBuilder":
        if self._cluster_name:
            raise AttributeError("'DropQuery' object already has attribute cluster_name")
        self._cluster_name = cluster

    def get_sql(self, **kwargs: Any) -> str:
        query = super().get_sql(**kwargs)

        if self._drop_target_kind != "DICTIONARY" and self._cluster_name is not None:
            query += " ON CLUSTER " + format_quotes(self._cluster_name, super().QUOTE_CHAR)

        return query


class SQLLiteValueWrapper(ValueWrapper):
    def get_value_sql(self, **kwargs: Any) -> str:
        if isinstance(self.value, bool):
            return "1" if self.value else "0"
        return super().get_value_sql(**kwargs)


class SQLLiteQuery(Query):
    """
    Defines a query class for use with Microsoft SQL Server.
    """

    @classmethod
    def _builder(cls, **kwargs: Any) -> "SQLLiteQueryBuilder":
        return SQLLiteQueryBuilder(**kwargs)


class SQLLiteQueryBuilder(QueryBuilder):
    QUERY_CLS = SQLLiteQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=Dialects.SQLLITE, wrapper_cls=SQLLiteValueWrapper, **kwargs)
        self._insert_or_replace = False

    @builder
    def insert_or_replace(self, *terms: Any) -> "SQLLiteQueryBuilder":
        self._apply_terms(*terms)
        self._replace = True
        self._insert_or_replace = True

    def _replace_sql(self, **kwargs: Any) -> str:
        prefix = "INSERT OR " if self._insert_or_replace else ""
        return prefix + super()._replace_sql(**kwargs)



from unittest import TestCase

from pypika import (
    ClickHouseQuery,
    Database,
    Table,
)


class ClickHouseQueryTests(TestCase):
    def test_use_AS_keyword_for_alias(self):
        t = Table('abc')
        query = ClickHouseQuery.from_(t).select(t.foo.as_('f1'), t.bar.as_('f2'))
        self.assertEqual(str(query), 'SELECT "foo" AS "f1","bar" AS "f2" FROM "abc"')

    def test_use_SAMPLE_keyword(self):
        t = Table('abc')
        query = ClickHouseQuery.from_(t).select(t.foo).sample(10)
        self.assertEqual(str(query), 'SELECT "foo" FROM "abc" SAMPLE 10')

    def test_use_SAMPLE_with_offset_keyword(self):
        t = Table('abc')
        query = ClickHouseQuery.from_(t).select(t.foo).sample(10, 5)
        self.assertEqual(str(query), 'SELECT "foo" FROM "abc" SAMPLE 10 OFFSET 5')

    def test_use_FINAL_keyword(self):
        t = Table('abc')
        query = ClickHouseQuery.from_(t).select(t.foo).final()
        self.assertEqual(str(query), 'SELECT "foo" FROM "abc" FINAL')


class ClickHouseDeleteTests(TestCase):
    table_abc = Table("abc")

    def test_omit_where(self):
        q = ClickHouseQuery.from_("abc").delete()

        self.assertEqual('ALTER TABLE "abc" DELETE', str(q))

    def test_omit_where__table_schema(self):
        q = ClickHouseQuery.from_(Table("abc", "schema1")).delete()

        self.assertEqual('ALTER TABLE "schema1"."abc" DELETE', str(q))

    def test_where_field_equals(self):
        q1 = ClickHouseQuery.from_(self.table_abc).where(self.table_abc.foo == self.table_abc.bar).delete()
        q2 = ClickHouseQuery.from_(self.table_abc).where(self.table_abc.foo.eq(self.table_abc.bar)).delete()

        self.assertEqual('ALTER TABLE "abc" DELETE WHERE "foo"="bar"', str(q1))
        self.assertEqual('ALTER TABLE "abc" DELETE WHERE "foo"="bar"', str(q2))


class ClickHouseUpdateTests(TestCase):
    table_abc = Table("abc")

    def test_update(self):
        q = ClickHouseQuery.update(self.table_abc).where(self.table_abc.foo == 0).set("foo", "bar")

        self.assertEqual('ALTER TABLE "abc" UPDATE "foo"=\'bar\' WHERE "foo"=0', str(q))


class ClickHouseDropQuery(TestCase):
    table_abc = Table("abc")
    database_xyz = Database("mydb")
    cluster_name = "mycluster"

    def test_drop_database(self):
        q1 = ClickHouseQuery.drop_database(self.database_xyz)
        q2 = ClickHouseQuery.drop_database(self.database_xyz).on_cluster(self.cluster_name)
        q3 = ClickHouseQuery.drop_database(self.database_xyz).if_exists().on_cluster(self.cluster_name)

        self.assertEqual('DROP DATABASE "mydb"', str(q1))
        self.assertEqual('DROP DATABASE "mydb" ON CLUSTER "mycluster"', str(q2))
        self.assertEqual('DROP DATABASE IF EXISTS "mydb" ON CLUSTER "mycluster"', str(q3))

    def test_drop_table(self):
        q1 = ClickHouseQuery.drop_table(self.table_abc)
        q2 = ClickHouseQuery.drop_table(self.table_abc).on_cluster(self.cluster_name)
        q3 = ClickHouseQuery.drop_table(self.table_abc).if_exists().on_cluster(self.cluster_name)

        self.assertEqual('DROP TABLE "abc"', str(q1))
        self.assertEqual('DROP TABLE "abc" ON CLUSTER "mycluster"', str(q2))
        self.assertEqual('DROP TABLE IF EXISTS "abc" ON CLUSTER "mycluster"', str(q3))

    def test_drop_dictionary(self):
        q1 = ClickHouseQuery.drop_dictionary("dict")
        q2 = ClickHouseQuery.drop_dictionary("dict").on_cluster(self.cluster_name)
        q3 = ClickHouseQuery.drop_dictionary("dict").if_exists().on_cluster(self.cluster_name)

        self.assertEqual('DROP DICTIONARY "dict"', str(q1))
        self.assertEqual('DROP DICTIONARY "dict"', str(q2))  # NO CLUSTER
        self.assertEqual('DROP DICTIONARY IF EXISTS "dict"', str(q3))  # NO CLUSTER

    def test_drop_other(self):
        q1 = ClickHouseQuery.drop_quota("myquota")
        q2 = ClickHouseQuery.drop_user("myuser")
        q3 = ClickHouseQuery.drop_view("myview")

        self.assertEqual('DROP QUOTA "myquota"', str(q1))
        self.assertEqual('DROP USER "myuser"', str(q2))
        self.assertEqual('DROP VIEW "myview"', str(q3))


class DistinctOnTests(TestCase):
    table_abc = Table("abc")

    def test_distinct_on(self):
        q = ClickHouseQuery.from_(self.table_abc).distinct_on("lname", self.table_abc.fname).select("lname", "id")

        self.assertEqual('''SELECT DISTINCT ON("lname","fname") "lname","id" FROM "abc"''', str(q))


class LimitByTests(TestCase):
    table_abc = Table("abc")

    def test_limit_by(self):
        q = ClickHouseQuery.from_(self.table_abc).limit_by(1, "a", self.table_abc.b).select("a", "b", "c")

        self.assertEqual('''SELECT "a","b","c" FROM "abc" LIMIT 1 BY ("a","b")''', str(q))

    def test_limit_offset_by(self):
        q = ClickHouseQuery.from_(self.table_abc).limit_offset_by(1, 2, "a", self.table_abc.b).select("a", "b", "c")

        self.assertEqual('''SELECT "a","b","c" FROM "abc" LIMIT 1 OFFSET 2 BY ("a","b")''', str(q))

    def test_limit_offset0_by(self):
        q = ClickHouseQuery.from_(self.table_abc).limit_offset_by(1, 0, "a", self.table_abc.b).select("a", "b", "c")

        self.assertEqual('''SELECT "a","b","c" FROM "abc" LIMIT 1 BY ("a","b")''', str(q))

    def test_rename_table(self):
        table_join = Table("join")

        q = (
            ClickHouseQuery.from_(self.table_abc)
            .join(table_join)
            .using("a")
            .limit_by(1, self.table_abc.a, table_join.a)
            .select(self.table_abc.b, table_join.b)
        )
        q = q.replace_table(self.table_abc, Table("xyz"))

        self.assertEqual(
            '''SELECT "xyz"."b","join"."b" FROM "xyz" JOIN "join" USING ("a") LIMIT 1 BY ("xyz"."a","join"."a")''',
            str(q),
        )



import unittest

from parameterized import parameterized

from pypika import Field
from pypika.clickhouse.array import (
    Array,
    Empty,
    HasAny,
    Length,
    NotEmpty,
)
from pypika.clickhouse.type_conversion import (
    ToFixedString,
    ToInt64,
)


class TestArray(unittest.TestCase):
    @parameterized.expand(
        [
            (
                "['ridley', 'scott', 'jimi', 'hendrix']",
                Array(["ridley", "scott", "jimi", "hendrix"]),
            ),
            (
                "[1, 2, 3, 4]",
                Array([1, 2, 3, 4]),
            ),
            (
                "[toInt64('1'),toInt64('2'),toInt64('3'),toInt64('4')]",
                Array(["1", "2", "3", "4"], ToInt64),
            ),
            (
                "[toFixedString('mogwai',10),toFixedString('mono',10),toFixedString('bonobo',10)] arr",
                Array(["mogwai", "mono", "bonobo"], ToFixedString, {"length": 10}).as_('arr'),
            ),
        ],
    )
    def test_get_sql(self, expected: str, array: Array):
        self.assertEqual(expected, array.get_sql())


class TestHasAny(unittest.TestCase):
    @parameterized.expand(
        [
            (
                'hasAny("mental_abilities","physical_abilities")',
                HasAny(Field("mental_abilities"), Field("physical_abilities")),
            ),
            ("hasAny([1, 2, 3, 4],[3])", HasAny(Array([1, 2, 3, 4]), Array([3]))),
            (
                "hasAny(\"bands\",[toFixedString('port-royal',20),toFixedString('hammock',20)])",
                HasAny(
                    Field("bands"),
                    Array(["port-royal", "hammock"], ToFixedString, {"length": 20}),
                ),
            ),
        ],
    )
    def test_get_sql(self, expected: str, func: HasAny):
        self.assertEqual(expected, func.get_sql())


class TestLength(unittest.TestCase):
    @parameterized.expand(
        [
            (
                'length("tags")',
                Length(Field("tags")),
            ),
            ("length([1, 2, 3])", Length(Array([1, 2, 3]))),
        ],
    )
    def test_get_sql(self, expected: str, func: Length):
        self.assertEqual(expected, func.get_sql())


class TestEmpty(unittest.TestCase):
    @parameterized.expand(
        [
            (
                'empty("tags")',
                Empty(Field("tags")),
            ),
            ("empty([1, 2, 3])", Empty(Array([1, 2, 3]))),
        ],
    )
    def test_get_sql(self, expected: str, func: Empty):
        self.assertEqual(expected, func.get_sql())


class TestNotEmpty(unittest.TestCase):
    @parameterized.expand(
        [
            (
                'notEmpty("tags")',
                NotEmpty(Field("tags")),
            ),
            ("notEmpty([1, 2, 3])", NotEmpty(Array([1, 2, 3]))),
        ],
    )
    def test_get_sql(self, expected: str, func: NotEmpty):
        self.assertEqual(expected, func.get_sql())



import unittest

from pypika import (
    Column,
    Tables,
)
from pypika import (
    functions as fn,
)
from pypika.dialects import SnowflakeQuery


class QuoteTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_use_double_quotes_on_alias_but_not_on_terms_1(self):
        q = SnowflakeQuery.from_(self.table_abc).select(self.table_abc.a.as_("bar"))

        self.assertEqual('SELECT a "bar" FROM abc', q.get_sql(with_namespace=True))

    def test_use_double_quotes_on_alias_but_not_on_terms_with_joins(self):
        foo = self.table_abc.as_("foo")
        bar = self.table_efg.as_("bar")

        q = SnowflakeQuery.from_(foo).join(bar).on(foo.fk == bar.id).select(foo.a, bar.b)

        self.assertEqual(
            "SELECT foo.a,bar.b FROM abc \"foo\" JOIN efg \"bar\" ON foo.fk=bar.id",
            q.get_sql(with_namespace=True),
        )

    def test_use_double_quotes_on_alias_but_not_on_terms_2(self):
        idx = self.table_abc.index.as_("idx")
        val = fn.Sum(self.table_abc.value).as_("val")
        q = SnowflakeQuery.from_(self.table_abc).select(idx, val).groupby(idx).orderby(idx)

        self.assertEqual(
            'SELECT index "idx",SUM(value) "val" FROM abc GROUP BY "idx" ORDER BY "idx"',
            q.get_sql(with_namespace=True),
        )

    def test_dont_use_double_quotes_on_joining_queries(self):
        foo = self.table_abc
        bar = self.table_efg
        q1 = SnowflakeQuery.from_(foo).select(foo.b)
        q2 = SnowflakeQuery.from_(bar).select(bar.b)
        q = SnowflakeQuery.from_(q1).join(q2).on(q1.b == q2.b).select("*")

        self.assertEqual(
            "SELECT * FROM (SELECT b FROM abc) sq0 JOIN (SELECT b FROM efg) sq1 ON sq0.b=sq1.b",
            q.get_sql(),
        )

    def test_dont_use_double_quotes_on_create_queries(self):
        q = SnowflakeQuery.create_table(self.table_abc).columns(Column("id", "INT"))
        self.assertEqual("CREATE TABLE abc (id INT)", q.get_sql())

    def test_dont_use_double_quotes_on_drop_queries(self):
        q = SnowflakeQuery.drop_table(self.table_abc)
        self.assertEqual("DROP TABLE abc", q.get_sql())



import unittest

from parameterized import parameterized

from pypika import ClickHouseQuery, Field, Table
from pypika.clickhouse.type_conversion import (
    ToDate,
    ToDateTime,
    ToFixedString,
    ToFloat32,
    ToFloat64,
    ToInt8,
    ToInt16,
    ToInt32,
    ToInt64,
    ToString,
    ToUInt8,
    ToUInt16,
    ToUInt32,
    ToUInt64,
)


class TestBasicTypeConverters(unittest.TestCase):
    @parameterized.expand(
        [
            (
                'toString("field_name")',
                ToString(Field("field_name")),
            ),
            (
                'toInt8("field_name")',
                ToInt8(Field("field_name")),
            ),
            (
                'toInt16("field_name")',
                ToInt16(Field("field_name")),
            ),
            (
                'toInt32("field_name")',
                ToInt32(Field("field_name")),
            ),
            (
                'toInt64("field_name")',
                ToInt64(Field("field_name")),
            ),
            (
                'toUInt8("field_name")',
                ToUInt8(Field("field_name")),
            ),
            (
                'toUInt16("field_name")',
                ToUInt16(Field("field_name")),
            ),
            (
                'toUInt32("field_name")',
                ToUInt32(Field("field_name")),
            ),
            (
                'toUInt64("field_name")',
                ToUInt64(Field("field_name")),
            ),
            (
                'toFloat32("field_name")',
                ToFloat32(Field("field_name")),
            ),
            (
                'toFloat64("field_name")',
                ToFloat64(Field("field_name")),
            ),
            (
                'toFloat64("field_name")',
                ToFloat64(Field("field_name")),
            ),
            (
                'toDate("field_name")',
                ToDate(Field("field_name")),
            ),
            (
                'toDateTime("field_name")',
                ToDateTime(Field("field_name")),
            ),
            (
                'toFixedString("field_name",100)',
                ToFixedString(Field("field_name"), 100),
            ),
        ],
    )
    def test_basic_types_field(self, expected, func):
        self.assertEqual(func, expected)

    @parameterized.expand(
        [
            (
                "toString('100')",
                ToString("100"),
            ),
            (
                "toInt8('100')",
                ToInt8("100"),
            ),
            (
                "toInt16('100')",
                ToInt16("100"),
            ),
            (
                "toInt32('100')",
                ToInt32("100"),
            ),
            (
                "toInt64('100')",
                ToInt64("100"),
            ),
            (
                "toUInt8('100')",
                ToUInt8("100"),
            ),
            (
                "toUInt16('100')",
                ToUInt16("100"),
            ),
            (
                "toUInt32('100')",
                ToUInt32("100"),
            ),
            (
                "toUInt64('100')",
                ToUInt64("100"),
            ),
            (
                "toFloat32('100')",
                ToFloat32("100"),
            ),
            (
                "toFloat64('100')",
                ToFloat64("100"),
            ),
            (
                "toFloat64('100')",
                ToFloat64("100"),
            ),
            (
                "toDate('100')",
                ToDate("100"),
            ),
            (
                "toDateTime('100')",
                ToDateTime("100"),
            ),
            (
                "toFixedString('100',100)",
                ToFixedString("100", 100),
            ),
        ],
    )
    def test_basic_types_value(self, expected, func):
        self.assertEqual(func, expected)


class TestToFixedString(unittest.TestCase):
    def test_get_sql_with_table(self):
        table = Table("example")
        query = (
            ClickHouseQuery.from_(table)
            .select(
                table.name,
            )
            .where(
                table.name == ToFixedString("name", 50),
            )
        )

        self.assertEqual(
            'SELECT "name" FROM "example" WHERE "name"=toFixedString(\'name\',50)',
            query.get_sql(),
        )



import unittest

from pypika import Query, Rollup, RollupException, Table
from pypika import functions as fn

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class RollupTests(unittest.TestCase):
    table = Table("abc")

    def test_mysql_one_groupby(self):
        q = (
            Query.from_(self.table)
            .select(self.table.foo, fn.Sum(self.table.bar))
            .rollup(self.table.foo, vendor="mysql")
        )

        self.assertEqual('SELECT "foo",SUM("bar") FROM "abc" GROUP BY "foo" WITH ROLLUP', str(q))

    def test_mysql_rollup_two_groupbys(self):
        q = (
            Query.from_(self.table)
            .select(self.table.foo, self.table.fiz, fn.Sum(self.table.bar))
            .rollup(self.table.foo, self.table.fiz, vendor="mysql")
        )

        self.assertEqual(
            'SELECT "foo","fiz",SUM("bar") FROM "abc" GROUP BY "foo","fiz" WITH ROLLUP',
            str(q),
        )

    def test_no_rollup_before_groupby(self):
        with self.assertRaises(RollupException):
            Query.from_(self.table).select(self.table.foo, fn.Sum(self.table.bar)).rollup(vendor="mysql")

    def test_no_rollup_after_rollup_mysql(self):
        with self.assertRaises(AttributeError):
            Query.from_(self.table).select(self.table.foo, self.table.fiz, fn.Sum(self.table.bar)).rollup(
                self.table.foo,
                vendor="mysql",
            ).rollup(self.table.fiz, vendor="mysql")

    def test_verticaoracle_func_one_groupby(self):
        q = Query.from_(self.table).select(self.table.foo, fn.Sum(self.table.bar)).groupby(Rollup(self.table.foo))

        self.assertEqual('SELECT "foo",SUM("bar") FROM "abc" GROUP BY ROLLUP("foo")', str(q))

    def test_verticaoracle_func_two_groupbys(self):
        q = (
            Query.from_(self.table)
            .select(self.table.foo, self.table.fiz, fn.Sum(self.table.bar))
            .groupby(
                Rollup(
                    self.table.foo,
                    self.table.fiz,
                ),
            )
        )

        self.assertEqual(
            'SELECT "foo","fiz",SUM("bar") FROM "abc" GROUP BY ROLLUP("foo","fiz")',
            str(q),
        )

    def test_verticaoracle_func_partial(self):
        q = (
            Query.from_(self.table)
            .select(self.table.foo, self.table.fiz, self.table.buz, fn.Sum(self.table.bar))
            .groupby(
                Rollup(
                    self.table.foo,
                    self.table.fiz,
                ),
                self.table.buz,
            )
        )

        self.assertEqual(
            'SELECT "foo","fiz","buz",SUM("bar") FROM "abc" GROUP BY ROLLUP("foo","fiz"),"buz"',
            str(q),
        )

    def test_verticaoracle_from_groupbys(self):
        q = Query.from_(self.table).select(self.table.foo, fn.Sum(self.table.bar)).rollup(self.table.foo)

        self.assertEqual('SELECT "foo",SUM("bar") FROM "abc" GROUP BY ROLLUP("foo")', str(q))

    def test_verticaoracle_from_two_groupbys(self):
        q = (
            Query.from_(self.table)
            .select(self.table.foo, self.table.fiz, fn.Sum(self.table.bar))
            .rollup(
                self.table.foo,
                self.table.fiz,
            )
        )

        self.assertEqual(
            'SELECT "foo","fiz",SUM("bar") FROM "abc" GROUP BY ROLLUP("foo","fiz")',
            str(q),
        )

    def test_verticaoracle_from_parameters(self):
        q = (
            Query.from_(self.table)
            .select(self.table.foo, self.table.fiz, fn.Sum(self.table.bar))
            .groupby(
                self.table.foo,
            )
            .rollup(
                self.table.fiz,
            )
        )

        self.assertEqual(
            'SELECT "foo","fiz",SUM("bar") FROM "abc" GROUP BY "foo",ROLLUP("fiz")',
            str(q),
        )

    def test_verticaoracle_multiple_rollups(self):
        q = (
            Query.from_(self.table)
            .select(self.table.foo, self.table.fiz, fn.Sum(self.table.bar))
            .rollup(
                self.table.foo,
            )
            .rollup(
                self.table.fiz,
            )
        )

        self.assertEqual(
            'SELECT "foo","fiz",SUM("bar") FROM "abc" GROUP BY ROLLUP("foo","fiz")',
            str(q),
        )

    def test_verticaoracle_rollups_with_parity(self):
        q = (
            Query.from_(self.table)
            .select(
                self.table.buz,
            )
            .rollup(
                [self.table.foo, self.table.bar],
                self.table.fiz,
            )
        )

        self.assertEqual('SELECT "buz" FROM "abc" GROUP BY ROLLUP(("foo","bar"),"fiz")', str(q))

    def test_verticaoracle_rollups_with_multiple_rollups_and_parity(self):
        q = (
            Query.from_(self.table)
            .select(
                self.table.buz,
            )
            .rollup(
                [self.table.foo, self.table.bar],
            )
            .rollup(
                [self.table.fiz, self.table.buz],
            )
        )

        self.assertEqual(
            'SELECT "buz" FROM "abc" GROUP BY ROLLUP(("foo","bar"),("fiz","buz"))',
            str(q),
        )



import unittest

from pypika import Table
from pypika.dialects import SQLLiteQuery


class SelectTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_bool_true_as_one(self):
        q = SQLLiteQuery.from_("abc").select(True)

        self.assertEqual('SELECT 1 FROM "abc"', str(q))

    def test_bool_false_as_zero(self):
        q = SQLLiteQuery.from_("abc").select(False)

        self.assertEqual('SELECT 0 FROM "abc"', str(q))


class ReplaceTests(unittest.TestCase):
    def test_normal_replace(self):
        query = SQLLiteQuery.into("abc").replace("v1", "v2", "v3")
        expected_output = "REPLACE INTO \"abc\" VALUES ('v1','v2','v3')"
        self.assertEqual(expected_output, str(query))

    def test_replace_subquery(self):
        query = SQLLiteQuery.into("abc").replace(SQLLiteQuery.from_("def").select("f1", "f2"))
        expected_output = 'REPLACE INTO "abc" VALUES ((SELECT "f1","f2" FROM "def"))'
        self.assertEqual(expected_output, str(query))

    def test_insert_or_replace(self):
        query = SQLLiteQuery.into("abc").insert_or_replace("v1", "v2", "v3")
        expected_output = "INSERT OR REPLACE INTO \"abc\" VALUES ('v1','v2','v3')"
        self.assertEqual(expected_output, str(query))



"""
PyPika is divided into a couple of modules, primarily the ``queries`` and ``terms`` modules.

pypika.analytics
----------------

Wrappers for SQL analytic functions

pypika.dialects
---------------

This contains all of the dialect specific implementations of the ``Query`` class.

pypika.enums
------------

Enumerated values are kept in this package which are used as options for Queries and Terms.

pypika.functions
----------------

Wrappers for common SQL functions are stored in this package.

pypika.pseudocolumns
--------------------

Wrappers for common SQL pseudocolumns are stored in this package.

pypika.queries
--------------

This is where the ``Query`` class can be found which is the core class in PyPika.  Also, other top level classes such
as ``Table`` can be found here.  ``Query`` is a container that holds all of the ``Term`` types together and also
serializes the builder to a string.

pypika.terms
------------

This module contains the classes which represent individual parts of queries that extend the ``Term`` base class.

pypika.utils
------------

This contains all of the utility classes such as exceptions and decorators.
"""

# noinspection PyUnresolvedReferences
from pypika.dialects import (
    ClickHouseQuery,
    Dialects,
    MSSQLQuery,
    MySQLQuery,
    OracleQuery,
    PostgreSQLQuery,
    RedshiftQuery,
    SQLLiteQuery,
    VerticaQuery,
)

# noinspection PyUnresolvedReferences
from pypika.enums import (
    DatePart,
    JoinType,
    Order,
)

# noinspection PyUnresolvedReferences
from pypika.queries import (
    AliasedQuery,
    Column,
    Database,
    Query,
    Schema,
    Table,
    Values,
    ValuesTuple,
)
from pypika.queries import (
    make_columns as Columns,
)
from pypika.queries import (
    make_tables as Tables,
)

# noinspection PyUnresolvedReferences
from pypika.terms import (
    JSON,
    Array,
    Bracket,
    Case,
    Criterion,
    CustomFunction,
    EmptyCriterion,
    Field,
    FormatParameter,
    Index,
    Interval,
    NamedParameter,
    Not,
    NullValue,
    NumericParameter,
    Parameter,
    PyformatParameter,
    QmarkParameter,
    Rollup,
    SystemTimeValue,
    Tuple,
)

# noinspection PyUnresolvedReferences
from pypika.utils import (
    CaseException,
    FunctionException,
    GroupingException,
    JoinException,
    QueryException,
    RollupException,
    SetOperationException,
)

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"
__version__ = "0.51.0"

NULL = NullValue()
SYSTEM_TIME = SystemTimeValue()

__all__ = (
    'JSON',
    'NULL',
    'SYSTEM_TIME',
    'AliasedQuery',
    'Array',
    'Bracket',
    'Case',
    'CaseException',
    'ClickHouseQuery',
    'Column',
    'Columns',
    'Criterion',
    'CustomFunction',
    'Database',
    'DatePart',
    'Dialects',
    'EmptyCriterion',
    'Field',
    'FormatParameter',
    'FunctionException',
    'GroupingException',
    'Index',
    'Interval',
    'JoinException',
    'JoinType',
    'MSSQLQuery',
    'MySQLQuery',
    'NamedParameter',
    'Not',
    'NullValue',
    'NumericParameter',
    'OracleQuery',
    'Order',
    'Parameter',
    'PostgreSQLQuery',
    'PyformatParameter',
    'QmarkParameter',
    'Query',
    'QueryException',
    'RedshiftQuery',
    'Rollup',
    'RollupException',
    'SQLLiteQuery',
    'Schema',
    'SetOperationException',
    'SystemTimeValue',
    'Table',
    'Tables',
    'Tuple',
    'Values',
    'VerticaQuery',
    'ValuesTuple',
)



import unittest

from pypika import Query, Tables
from pypika import functions as fn


class QuoteTests(unittest.TestCase):
    maxDiff = None

    table_abc, table_efg = Tables("abc", "efg")

    def setUp(self):
        subquery1 = (
            Query.from_(self.table_abc)
            .select(
                self.table_abc.foo,
                fn.Sum(self.table_abc.fizz + self.table_abc.buzz).as_("fizzbuzz"),
            )
            .groupby(self.table_abc.foo)
        )

        subquery2 = Query.from_(self.table_efg).select(
            self.table_efg.foo.as_("foo_two"),
            self.table_efg.bar,
        )

        self.query = (
            Query.from_(subquery1)
            .select(subquery1.foo, subquery1.fizzbuzz)
            .join(subquery2)
            .on(subquery1.foo == subquery2.foo_two)
            .select(subquery2.foo_two, subquery2.bar)
        )

    def test_replace_quote_char_in_complex_query(self):
        self.assertEqual(
            "SELECT "
            "`sq0`.`foo`,`sq0`.`fizzbuzz`,"
            "`sq1`.`foo_two`,`sq1`.`bar` "
            "FROM ("
            "SELECT "
            "`foo`,SUM(`fizz`+`buzz`) `fizzbuzz` "
            "FROM `abc` "
            "GROUP BY `foo`"
            ") `sq0` JOIN ("
            "SELECT "
            "`foo` `foo_two`,`bar` "
            "FROM `efg`"
            ") `sq1` ON `sq0`.`foo`=`sq1`.`foo_two`",
            self.query.get_sql(quote_char="`"),
        )

    def test_no_quote_char_in_complex_query(self):
        self.assertEqual(
            "SELECT "
            "sq0.foo,sq0.fizzbuzz,"
            "sq1.foo_two,sq1.bar "
            "FROM ("
            "SELECT "
            "foo,SUM(fizz+buzz) fizzbuzz "
            "FROM abc "
            "GROUP BY foo"
            ") sq0 JOIN ("
            "SELECT "
            "foo foo_two,bar "
            "FROM efg"
            ") sq1 ON sq0.foo=sq1.foo_two",
            self.query.get_sql(quote_char=None),
        )



import unittest

from pypika import SYSTEM_TIME, AliasedQuery, PostgreSQLQuery, Query, SQLLiteQuery, Table, Values, ValuesTuple

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"

from pypika.terms import Star


class UpdateTests(unittest.TestCase):
    table_abc = Table("abc")
    table_def = Table("def")

    def test_empty_query(self):
        q = Query.update("abc")

        self.assertEqual("", str(q))

    def test_omit_where(self):
        q = Query.update(self.table_abc).set("foo", "bar")

        self.assertEqual('UPDATE "abc" SET "foo"=\'bar\'', str(q))

    def test_single_quote_escape_in_set(self):
        q = Query.update(self.table_abc).set("foo", "bar'foo")

        self.assertEqual("UPDATE \"abc\" SET \"foo\"='bar''foo'", str(q))

    def test_update__table_schema(self):
        table = Table("abc", "schema1")
        q = Query.update(table).set(table.foo, 1).where(table.foo == 0)

        self.assertEqual('UPDATE "schema1"."abc" SET "foo"=1 WHERE "foo"=0', str(q))

    def test_update_with_none(self):
        q = Query.update("abc").set("foo", None)
        self.assertEqual('UPDATE "abc" SET "foo"=null', str(q))

    def test_update_with_join(self):
        q = (
            Query.update(self.table_abc)
            .join(self.table_def)
            .on(self.table_def.abc_id == self.table_abc.id)
            .set(self.table_abc.lname, self.table_def.lname)
        )
        self.assertEqual(
            'UPDATE "abc" JOIN "def" ON "def"."abc_id"="abc"."id" SET "lname"="def"."lname"',
            str(q),
        )

    def test_update_with_limit(self):
        q = Query.update(self.table_abc).set(self.table_abc.lname, "test").limit(1)
        self.assertEqual('UPDATE "abc" SET "lname"=\'test\' LIMIT 1', str(q))

    def test_update_from(self):
        from_table = Table('from_table')

        q = Query.update(self.table_abc).set(self.table_abc.lname, from_table.long_name).from_(from_table)
        self.assertEqual('UPDATE "abc" SET "lname"="from_table"."long_name" FROM "from_table"', str(q))

    def test_update_from_with_where(self):
        from_table = Table('from_table')

        q = (
            Query.update(self.table_abc)
            .set(self.table_abc.lname, from_table.long_name)
            .from_(from_table)
            .where(self.table_abc.fname.eq(from_table.full_name))
        )
        self.assertEqual(
            'UPDATE "abc" SET "lname"="from_table"."long_name" FROM "from_table" '
            'WHERE "abc"."fname"="from_table"."full_name"',
            str(q),
        )

    def test_update_with_statement(self):
        table_efg = Table('efg')

        sub_query = Query.from_(table_efg).select("fizz")
        an_alias = AliasedQuery("an_alias")

        q = (
            Query.with_(sub_query, "an_alias")
            .update(self.table_abc)
            .from_(an_alias)
            .set(self.table_abc.lname, an_alias.long_name)
            .where(self.table_abc.comp.eq(an_alias.alias_comp))
        )
        self.assertEqual(
            'WITH an_alias AS (SELECT "fizz" FROM "efg") '
            'UPDATE "abc" SET "lname"="an_alias"."long_name" FROM an_alias '
            'WHERE "abc"."comp"="an_alias"."alias_comp"',
            str(q),
        )

    def test_for_portion(self):
        with self.subTest("with system time"):
            q = Query.update(self.table_abc.for_portion(SYSTEM_TIME.from_to('2020-01-01', '2020-02-01'))).set(
                "foo",
                "bar",
            )

            self.assertEqual(
                'UPDATE "abc" FOR PORTION OF SYSTEM_TIME FROM \'2020-01-01\' TO \'2020-02-01\' SET "foo"=\'bar\'',
                str(q),
            )

        with self.subTest("with column"):
            q = Query.update(
                self.table_abc.for_portion(self.table_abc.valid_period.from_to('2020-01-01', '2020-02-01')),
            ).set("foo", "bar")

            self.assertEqual(
                'UPDATE "abc" FOR PORTION OF "valid_period" FROM \'2020-01-01\' TO \'2020-02-01\' SET "foo"=\'bar\'',
                str(q),
            )


class PostgresUpdateTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_update_returning_str(self):
        q = PostgreSQLQuery.update(self.table_abc).where(self.table_abc.foo == 0).set("foo", "bar").returning("id")

        self.assertEqual('UPDATE "abc" SET "foo"=\'bar\' WHERE "foo"=0 RETURNING "abc"."id"', str(q))

    def test_update_returning(self):
        q = (
            PostgreSQLQuery.update(self.table_abc)
            .where(self.table_abc.foo == 0)
            .set("foo", "bar")
            .returning(self.table_abc.id)
        )

        self.assertEqual('UPDATE "abc" SET "foo"=\'bar\' WHERE "foo"=0 RETURNING "abc"."id"', str(q))

    def test_update_returning_from_different_tables(self):
        table_bcd = Table('bcd')

        q = (
            PostgreSQLQuery.update(self.table_abc)
            .from_(table_bcd)
            .set(self.table_abc.lname, table_bcd.long_name)
            .returning(self.table_abc.id, table_bcd.fname)
        )
        self.assertEqual(
            'UPDATE "abc" SET "lname"="bcd"."long_name" FROM "bcd" RETURNING "abc"."id","bcd"."fname"',
            str(q),
        )

    def test_update_returning_star(self):
        q = PostgreSQLQuery.update(self.table_abc).where(self.table_abc.foo == 0).set("foo", "bar").returning(Star())

        self.assertEqual('UPDATE "abc" SET "foo"=\'bar\' WHERE "foo"=0 RETURNING *', str(q))

    def test_update_values(self):
        values = Values("data", ("id", "foo"), [(1, "bar"), (2, "bar")])

        q = (
            PostgreSQLQuery.update(self.table_abc)
            .from_(values)
            .where(self.table_abc.id == values.id)
            .set(self.table_abc.foo, values.foo)
        )

        self.assertEqual(
            'UPDATE "abc" SET "foo"="data"."foo" FROM (VALUES (1,\'bar\'),(2,\'bar\')) AS "data" ("id","foo") WHERE "abc"."id"="data"."id"',
            str(q),
        )

    def test_update_values_with_cast(self):
        values = ValuesTuple("data", ("id", "foo"), [(1, "bar"), (2, "bar")], as_type="abc")

        q = (
            PostgreSQLQuery.update(self.table_abc)
            .from_(values)
            .where(self.table_abc.id == values.id)
            .set(self.table_abc.foo, values.foo)
        )

        self.assertEqual(
            'UPDATE "abc" SET "foo"=("x")."foo" FROM (VALUES ((1,\'bar\')::abc),((2,\'bar\')::abc)) AS "data" ("x") WHERE "abc"."id"=("x")."id"',
            str(q),
        )


class SQLLiteUpdateTests(unittest.TestCase):
    table_abc = Table("abc")

    def test_update_with_bool(self):
        q = SQLLiteQuery.update(self.table_abc).set(self.table_abc.foo, True)

        self.assertEqual('UPDATE "abc" SET "foo"=1', str(q))



from pypika.terms import (
    Field,
    Function,
)
from pypika.utils import format_alias_sql


class ToString(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToString, self).__init__("toString", name, alias=alias)


class ToFixedString(Function):
    def __init__(self, field, length: int, alias: str | None = None, schema: str | None = None):
        self._length = length
        self._field = field
        self.alias = alias
        self.name = "toFixedString"
        self.schema = schema
        self.args = ()

    def get_sql(self, with_alias=False, with_namespace=False, quote_char=None, dialect=None, **kwargs):
        sql = "{name}({field},{length})".format(
            name=self.name,
            field=self._field if isinstance(self._field, Field) else "'%s'" % str(self._field),
            length=self._length,
        )
        return format_alias_sql(sql, self.alias, **kwargs)


class ToInt8(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToInt8, self).__init__("toInt8", name, alias=alias)


class ToInt16(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToInt16, self).__init__("toInt16", name, alias=alias)


class ToInt32(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToInt32, self).__init__("toInt32", name, alias=alias)


class ToInt64(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToInt64, self).__init__("toInt64", name, alias=alias)


class ToUInt8(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToUInt8, self).__init__("toUInt8", name, alias=alias)


class ToUInt16(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToUInt16, self).__init__("toUInt16", name, alias=alias)


class ToUInt32(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToUInt32, self).__init__("toUInt32", name, alias=alias)


class ToUInt64(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToUInt64, self).__init__("toUInt64", name, alias=alias)


class ToFloat32(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToFloat32, self).__init__("toFloat32", name, alias=alias)


class ToFloat64(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToFloat64, self).__init__("toFloat64", name, alias=alias)


class ToDate(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToDate, self).__init__("toDate", name, alias=alias)


class ToDateTime(Function):
    def __init__(self, name, alias: str | None = None):
        super(ToDateTime, self).__init__("toDateTime", name, alias=alias)



import unittest

from pypika import utils

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class ImmutabilityTests(unittest.TestCase):
    def test_raise_attribute_error_for_deepcopy(self):
        with self.assertRaises(AttributeError):
            utils.raise_attribute_for_copy("Test", "__deepcopy__")

    def test_raise_attribute_error_for_getstate(self):
        with self.assertRaises(AttributeError):
            utils.raise_attribute_for_copy("Test", "__getstate__")

    def test_raise_attribute_error_for_setstate(self):
        with self.assertRaises(AttributeError):
            utils.raise_attribute_for_copy("Test", "__setstate__")

    def test_raise_attribute_error_for_getnewargs(self):
        with self.assertRaises(AttributeError):
            utils.raise_attribute_for_copy("Test", "__getnewargs__")



import unittest

from pypika import Criterion, JoinType, Order, Query, Tables
from pypika import analytics as an
from pypika.analytics import Lag, Lead

__author__ = "Timothy Heys"
__email__ = "theys@kayak.com"


class RankTests(unittest.TestCase):
    table_abc, table_efg = Tables("abc", "efg")

    def test_rank(self):
        expr = an.Rank().over(self.table_abc.foo).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT RANK() OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_dense_rank(self):
        expr = an.DenseRank().over(self.table_abc.foo).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT DENSE_RANK() OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_row_number(self):
        expr = an.RowNumber().over(self.table_abc.foo).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT ROW_NUMBER() OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_rank_with_alias(self):
        expr = an.Rank().over(self.table_abc.foo).orderby(self.table_abc.date).as_("rank")

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT RANK() OVER(PARTITION BY \"foo\" ORDER BY \"date\") \"rank\" FROM \"abc\"",
            str(q),
        )

    def test_multiple_partitions(self):
        expr = an.Rank().orderby(self.table_abc.date).over(self.table_abc.foo, self.table_abc.bar)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT RANK() OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_ntile_no_partition_or_order_invalid_sql(self):
        expr = an.NTile(5)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual("SELECT NTILE(5) FROM \"abc\"", str(q))

    def test_ntile_with_partition(self):
        expr = an.NTile(5).over(self.table_abc.foo)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual("SELECT NTILE(5) OVER(PARTITION BY \"foo\") FROM \"abc\"", str(q))

    def test_ntile_with_order(self):
        expr = an.NTile(5).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual("SELECT NTILE(5) OVER(ORDER BY \"date\") FROM \"abc\"", str(q))

    def test_ntile_with_partition_and_order(self):
        expr = an.NTile(5).over(self.table_abc.foo).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT NTILE(5) OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_first_value(self):
        expr = an.FirstValue(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT FIRST_VALUE(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_first_value_ignore_nulls(self):
        expr = an.FirstValue(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).ignore_nulls()

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT FIRST_VALUE(\"fizz\" IGNORE NULLS) OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_first_value_ignore_nulls_first(self):
        expr = an.FirstValue(self.table_abc.fizz).ignore_nulls().over(self.table_abc.foo).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT FIRST_VALUE(\"fizz\" IGNORE NULLS) OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_first_value_multi_argument(self):
        expr = (
            an.FirstValue(self.table_abc.fizz, self.table_abc.buzz)
            .over(self.table_abc.foo)
            .orderby(self.table_abc.date)
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT FIRST_VALUE(\"fizz\",\"buzz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_last_value(self):
        expr = an.LastValue(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT LAST_VALUE(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_filter(self):
        expr = (
            an.LastValue(self.table_abc.fizz)
            .filter(Criterion.all([self.table_abc.bar == True]))
            .over(self.table_abc.foo)
            .orderby(self.table_abc.date)
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT "
            'LAST_VALUE("fizz") '
            'FILTER(WHERE "bar"=true) '
            'OVER(PARTITION BY "foo" ORDER BY "date") '
            'FROM "abc"',
            str(q),
        )

    def test_filter_quote_table_in_filter(self):
        expr = (
            an.LastValue(self.table_efg.fizz)
            .filter(self.table_efg.filed.eq('yes'))
            .over(self.table_efg.foo)
            .orderby(self.table_efg.date)
        )

        q = (
            Query.from_(self.table_abc)
            .inner_join(self.table_efg)
            .on(self.table_abc.id.eq(self.table_efg.id))
            .select(expr)
        )

        self.assertEqual(
            'SELECT '
            'LAST_VALUE("efg"."fizz") '
            'FILTER(WHERE "efg"."filed"=\'yes\') '
            'OVER(PARTITION BY "efg"."foo" ORDER BY "efg"."date") '
            'FROM "abc" '
            'JOIN "efg" ON "abc"."id"="efg"."id"',
            str(q),
        )

    def test_orderby_asc(self):
        expr = an.LastValue(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date, order=Order.asc)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT LAST_VALUE(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\" ASC) FROM \"abc\"",
            str(q),
        )

    def test_orderby_desc(self):
        expr = an.LastValue(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date, order=Order.desc)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT LAST_VALUE(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\" DESC) FROM \"abc\"",
            str(q),
        )

    def test_last_value_multi_argument(self):
        expr = (
            an.LastValue(self.table_abc.fizz, self.table_abc.buzz).over(self.table_abc.foo).orderby(self.table_abc.date)
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT LAST_VALUE(\"fizz\",\"buzz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_last_value_ignore_nulls(self):
        expr = an.LastValue(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).ignore_nulls()

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT LAST_VALUE(\"fizz\" IGNORE NULLS) OVER(PARTITION BY \"foo\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_median(self):
        expr = an.Median(self.table_abc.fizz).over(self.table_abc.foo, self.table_abc.bar).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT MEDIAN(\"fizz\") OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_avg(self):
        expr = an.Avg(self.table_abc.fizz).over(self.table_abc.foo, self.table_abc.bar).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT AVG(\"fizz\") OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_stddev(self):
        expr = an.StdDev(self.table_abc.fizz).over(self.table_abc.foo, self.table_abc.bar).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT STDDEV(\"fizz\") OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_stddev_pop(self):
        expr = (
            an.StdDevPop(self.table_abc.fizz).over(self.table_abc.foo, self.table_abc.bar).orderby(self.table_abc.date)
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT STDDEV_POP(\"fizz\") OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_stddev_samp(self):
        expr = (
            an.StdDevSamp(self.table_abc.fizz).over(self.table_abc.foo, self.table_abc.bar).orderby(self.table_abc.date)
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT STDDEV_SAMP(\"fizz\") OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_variance(self):
        expr = (
            an.Variance(self.table_abc.fizz).over(self.table_abc.foo, self.table_abc.bar).orderby(self.table_abc.date)
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT VARIANCE(\"fizz\") OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_var_pop(self):
        expr = an.VarPop(self.table_abc.fizz).over(self.table_abc.foo, self.table_abc.bar).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT VAR_POP(\"fizz\") OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_var_samp(self):
        expr = an.VarSamp(self.table_abc.fizz).over(self.table_abc.foo, self.table_abc.bar).orderby(self.table_abc.date)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT VAR_SAMP(\"fizz\") OVER(PARTITION BY \"foo\",\"bar\" ORDER BY \"date\") FROM \"abc\"",
            str(q),
        )

    def test_table_prefixes_used_in_analytic_functions(self):
        expr = an.Rank().over(self.table_abc.foo).orderby(self.table_efg.date)

        query = (
            Query.from_(self.table_abc)
            .join(self.table_efg, how=JoinType.left)
            .on(self.table_abc.foo == self.table_efg.bar)
            .select("*", expr)
        )

        self.assertEqual(
            'SELECT *,RANK() OVER(PARTITION BY "abc"."foo" ORDER BY "efg"."date") '
            'FROM "abc" LEFT JOIN "efg" ON "abc"."foo"="efg"."bar"',
            str(query),
        )

    def test_sum_rows_unbounded_preceeding(self):
        expr = an.Sum(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).rows(an.Preceding())

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT SUM(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\" ROWS UNBOUNDED PRECEDING) FROM \"abc\"",
            str(q),
        )

    def test_max_rows_x_preceeding(self):
        expr = an.Max(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).rows(an.Preceding(5))

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT MAX(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\" ROWS 5 PRECEDING) FROM \"abc\"",
            str(q),
        )

    def test_min_rows_current_row(self):
        expr = an.Min(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).rows(an.CURRENT_ROW)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT MIN(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\" ROWS CURRENT ROW) FROM \"abc\"",
            str(q),
        )

    def test_varpop_range_unbounded_preceeding(self):
        expr = (
            an.VarPop(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).range(an.Preceding())
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT "
            'VAR_POP("fizz") '
            "OVER("
            'PARTITION BY "foo" ORDER BY "date" '
            "RANGE UNBOUNDED PRECEDING"
            ") "
            'FROM "abc"',
            str(q),
        )

    def test_max_range_x_preceeding(self):
        expr = an.Max(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).range(an.Preceding(5))

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT MAX(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\" RANGE 5 PRECEDING) FROM \"abc\"",
            str(q),
        )

    def test_min_range_current_row(self):
        expr = an.Min(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).range(an.CURRENT_ROW)

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT MIN(\"fizz\") OVER(PARTITION BY \"foo\" ORDER BY \"date\" RANGE CURRENT ROW) FROM \"abc\"",
            str(q),
        )

    def test_variance_rows_between_unbounded_preceeding_unbounded_following(self):
        expr = (
            an.Variance(self.table_abc.fizz)
            .over(self.table_abc.foo)
            .orderby(self.table_abc.date)
            .rows(an.Preceding(), an.Following())
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT "
            'VARIANCE("fizz") '
            "OVER("
            'PARTITION BY "foo" ORDER BY "date" '
            "ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING"
            ") "
            'FROM "abc"',
            str(q),
        )

    def test_first_value_range_between_x_preceeding_unbounded_following(self):
        expr = (
            an.FirstValue(self.table_abc.fizz)
            .over(self.table_abc.foo)
            .orderby(self.table_abc.date)
            .range(an.Preceding(3), an.Following())
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT "
            'FIRST_VALUE("fizz") '
            "OVER("
            'PARTITION BY "foo" ORDER BY "date" '
            "RANGE BETWEEN 3 PRECEDING AND UNBOUNDED FOLLOWING"
            ") "
            'FROM "abc"',
            str(q),
        )

    def test_varpop_rows_between_unbounded_preceeding_x_following(self):
        expr = (
            an.VarPop(self.table_abc.fizz)
            .over(self.table_abc.foo)
            .orderby(self.table_abc.date)
            .rows(an.Preceding(), an.Following(6))
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT "
            'VAR_POP("fizz") '
            "OVER("
            'PARTITION BY "foo" ORDER BY "date" '
            "ROWS BETWEEN UNBOUNDED PRECEDING AND 6 FOLLOWING"
            ") "
            'FROM "abc"',
            str(q),
        )

    def test_count_range_between_unbounded_preceeding_current_row(self):
        expr = (
            an.Count(self.table_abc.fizz)
            .over(self.table_abc.foo)
            .orderby(self.table_abc.date)
            .range(an.Preceding(), an.CURRENT_ROW)
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT "
            'COUNT("fizz") '
            "OVER("
            'PARTITION BY "foo" ORDER BY "date" '
            "RANGE BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW"
            ") "
            'FROM "abc"',
            str(q),
        )

    def test_last_value_rows_between_current_row_unbounded_following(self):
        expr = (
            an.LastValue(self.table_abc.fizz)
            .over(self.table_abc.foo)
            .orderby(self.table_abc.date)
            .rows(an.CURRENT_ROW, an.Following())
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT "
            'LAST_VALUE("fizz") '
            "OVER("
            'PARTITION BY "foo" ORDER BY "date" '
            "ROWS BETWEEN CURRENT ROW AND UNBOUNDED FOLLOWING"
            ") "
            'FROM "abc"',
            str(q),
        )

    def test_last_value_rows_between_current_row_unbounded_following_ignore_nulls(self):
        expr = (
            an.LastValue(self.table_abc.fizz)
            .over(self.table_abc.foo)
            .orderby(self.table_abc.date)
            .ignore_nulls()
            .rows(an.CURRENT_ROW, an.Following(8))
        )

        q = Query.from_(self.table_abc).select(expr)

        self.assertEqual(
            "SELECT "
            'LAST_VALUE("fizz" IGNORE NULLS) '
            "OVER("
            'PARTITION BY "foo" ORDER BY "date" '
            "ROWS BETWEEN CURRENT ROW AND 8 FOLLOWING"
            ") "
            'FROM "abc"',
            str(q),
        )

    def test_empty_over(self):
        query = Query.from_(self.table_abc).select(an.Sum(self.table_abc.fizz).over())

        self.assertEqual("SELECT SUM(\"fizz\") OVER() FROM \"abc\"", str(query))

    def test_rows_called_twice_raises_attribute_error(self):
        with self.assertRaises(AttributeError):
            an.Sum(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).rows(an.Preceding()).rows(
                an.Preceding(),
            )

    def test_range_called_twice_raises_attribute_error(self):
        with self.assertRaises(AttributeError):
            an.Sum(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).range(
                an.Preceding(),
            ).range(an.Preceding())

    def test_rows_then_range_raises_attribute_error(self):
        with self.assertRaises(AttributeError):
            an.Sum(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).rows(
                an.Preceding(),
            ).range(an.Preceding())

    def test_range_then_rows_raises_attribute_error(self):
        with self.assertRaises(AttributeError):
            an.Sum(self.table_abc.fizz).over(self.table_abc.foo).orderby(self.table_abc.date).range(
                an.Preceding(),
            ).rows(an.Preceding())

    def test_lag_generates_correct_sql(self):
        with self.subTest('without partition by'):
            q = Query.from_(self.table_abc).select(
                Lag(self.table_abc.date, 1, '2000-01-01').orderby(self.table_abc.date),
            )

            self.assertEqual('SELECT LAG("date",1,\'2000-01-01\') OVER(ORDER BY "date") FROM "abc"', str(q))

        with self.subTest('with partition by'):
            q = Query.from_(self.table_abc).select(
                Lag(self.table_abc.date, 1, '2000-01-01').over(self.table_abc.foo).orderby(self.table_abc.date),
            )

            self.assertEqual(
                'SELECT LAG("date",1,\'2000-01-01\') OVER(PARTITION BY "foo" ORDER BY "date") FROM "abc"',
                str(q),
            )

    def test_lead_generates_correct_sql(self):
        with self.subTest('without partition by'):
            q = Query.from_(self.table_abc).select(
                Lead(self.table_abc.date, 1, '2000-01-01').orderby(self.table_abc.date),
            )

            self.assertEqual('SELECT LEAD("date",1,\'2000-01-01\') OVER(ORDER BY "date") FROM "abc"', str(q))

        with self.subTest('with partition by'):
            q = Query.from_(self.table_abc).select(
                Lead(self.table_abc.date, 1, '2000-01-01').over(self.table_abc.foo).orderby(self.table_abc.date),
            )

            self.assertEqual(
                'SELECT LEAD("date",1,\'2000-01-01\') OVER(PARTITION BY "foo" ORDER BY "date") FROM "abc"',
                str(q),
            )



import unittest

from pypika import (
    Field as F,
)
from pypika import (
    Interval,
)
from pypika.enums import Dialects

dt = F("dt")


class AddIntervalTests(unittest.TestCase):
    def test_add_microsecond(self):
        c = dt + Interval(microseconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1 MICROSECOND'", str(c))

    def test_add_second(self):
        c = dt + Interval(seconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1 SECOND'", str(c))

    def test_add_minute(self):
        c = dt + Interval(minutes=1)

        self.assertEqual("\"dt\"+INTERVAL '1 MINUTE'", str(c))

    def test_add_day(self):
        c = dt + Interval(days=1)

        self.assertEqual("\"dt\"+INTERVAL '1 DAY'", str(c))

    def test_add_week(self):
        c = dt + Interval(weeks=1)

        self.assertEqual("\"dt\"+INTERVAL '1 WEEK'", str(c))

    def test_add_month(self):
        c = dt + Interval(months=1)

        self.assertEqual("\"dt\"+INTERVAL '1 MONTH'", str(c))

    def test_add_quarter(self):
        c = dt + Interval(quarters=1)

        self.assertEqual("\"dt\"+INTERVAL '1 QUARTER'", str(c))

    def test_add_year(self):
        c = dt + Interval(years=1)

        self.assertEqual("\"dt\"+INTERVAL '1 YEAR'", str(c))

    def test_add_default(self):
        c = dt + Interval(days=0)

        self.assertEqual("\"dt\"+INTERVAL '0 DAY'", str(c))


class AddIntervalMultipleUnitsTests(unittest.TestCase):
    def test_add_second_microsecond(self):
        c = dt + Interval(seconds=1, microseconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1.1 SECOND_MICROSECOND'", str(c))

    def test_add_minute_microsecond(self):
        c = dt + Interval(minutes=1, microseconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1:0.1 MINUTE_MICROSECOND'", str(c))

    def test_add_minute_second(self):
        c = dt + Interval(minutes=1, seconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1:1 MINUTE_SECOND'", str(c))

    def test_add_hour_microsecond(self):
        c = dt + Interval(hours=1, microseconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1:0:0.1 HOUR_MICROSECOND'", str(c))

    def test_add_hour_second(self):
        c = dt + Interval(hours=1, seconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1:0:1 HOUR_SECOND'", str(c))

    def test_add_hour_minute(self):
        c = dt + Interval(hours=1, minutes=1)

        self.assertEqual("\"dt\"+INTERVAL '1:1 HOUR_MINUTE'", str(c))

    def test_add_day_microsecond(self):
        c = dt + Interval(days=1, microseconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1 0:0:0.1 DAY_MICROSECOND'", str(c))

    def test_add_day_second(self):
        c = dt + Interval(days=1, seconds=1)

        self.assertEqual("\"dt\"+INTERVAL '1 0:0:1 DAY_SECOND'", str(c))

    def test_add_day_minute(self):
        c = dt + Interval(days=1, minutes=1)

        self.assertEqual("\"dt\"+INTERVAL '1 0:1 DAY_MINUTE'", str(c))

    def test_add_day_hour(self):
        c = dt + Interval(days=1, hours=1)

        self.assertEqual("\"dt\"+INTERVAL '1 1 DAY_HOUR'", str(c))

    def test_add_year_month(self):
        c = dt + Interval(years=1, months=1)

        self.assertEqual("\"dt\"+INTERVAL '1-1 YEAR_MONTH'", str(c))

    def test_add_value_right(self):
        c = Interval(microseconds=1) - dt

        self.assertEqual("INTERVAL '1 MICROSECOND'-\"dt\"", str(c))

    def test_add_value_complex_expressions(self):
        c = dt + Interval(quarters=1) + Interval(weeks=1)

        self.assertEqual("\"dt\"+INTERVAL '1 QUARTER'+INTERVAL '1 WEEK'", str(c))


class DialectIntervalTests(unittest.TestCase):
    def test_mysql_dialect_uses_single_quotes_around_expression_in_an_interval(self):
        c = Interval(days=1).get_sql(dialect=Dialects.MYSQL)
        self.assertEqual("INTERVAL '1' DAY", c)

    def test_oracle_dialect_uses_single_quotes_around_expression_in_an_interval(self):
        c = Interval(days=1).get_sql(dialect=Dialects.ORACLE)
        self.assertEqual("INTERVAL '1' DAY", c)

    def test_vertica_dialect_uses_single_quotes_around_interval(self):
        c = Interval(days=1).get_sql(dialect=Dialects.VERTICA)
        self.assertEqual("INTERVAL '1 DAY'", c)

    def test_redshift_dialect_uses_single_quotes_around_interval(self):
        c = Interval(days=1).get_sql(dialect=Dialects.REDSHIFT)
        self.assertEqual("INTERVAL '1 DAY'", c)

    def test_postgresql_dialect_uses_single_quotes_around_interval(self):
        c = Interval(days=1).get_sql(dialect=Dialects.POSTGRESQL)
        self.assertEqual("INTERVAL '1 DAY'", c)


class TestNegativeIntervals(unittest.TestCase):
    def test_day(self):
        c = Interval(days=-1).get_sql()
        self.assertEqual("INTERVAL '-1 DAY'", c)

    def test_week(self):
        c = Interval(weeks=-1).get_sql()
        self.assertEqual("INTERVAL '-1 WEEK'", c)

    def test_month(self):
        c = Interval(months=-1).get_sql()
        self.assertEqual("INTERVAL '-1 MONTH'", c)

    def test_year(self):
        c = Interval(years=-1).get_sql()
        self.assertEqual("INTERVAL '-1 YEAR'", c)

    def test_year_month(self):
        c = Interval(years=-1, months=-4).get_sql()
        self.assertEqual("INTERVAL '-1-4 YEAR_MONTH'", c)


class TruncateTrailingZerosTests(unittest.TestCase):
    def test_do_not_truncate_integer_values(self):
        i = Interval(seconds=10)

        self.assertEqual("INTERVAL '10 SECOND'", str(i))

    def test_do_not_truncate_months_between_years_and_datys(self):
        i = Interval(years=10, days=10)

        self.assertEqual("INTERVAL '10-0-10 YEAR_DAY'", str(i))
